import json
import logging
from typing import Dict, List, Optional, Union

import gspread

from .base import BaseOutput

try:
    import numpy
    from pandas import DataFrame
except ImportError:
    DataFrame = None


logger = logging.getLogger(__name__)


class GoogleSheets(BaseOutput):
    """GoogleSheets class enabling writing to Google spreadsheets.

    Args:
        configuration (Dict): Configuration for Google Sheets
        gsheet_auth (str): Authorisation for Google Sheets/Drive
        updatesheets (List[str]): List of spreadsheets to update (eg. prod, test)
        tabs (Dict[str, str]): Dictionary of mappings from internal name to spreadsheet tab name
        updatetabs (List[str]): Tabs to update
    """

    def __init__(
        self,
        configuration: Dict,
        gsheet_auth: str,
        updatesheets: List[str],
        tabs: Dict[str, str],
        updatetabs: List[str],
    ) -> None:
        super().__init__(updatetabs)
        info = json.loads(gsheet_auth)
        scopes = ["https://www.googleapis.com/auth/spreadsheets"]
        self.gc = gspread.service_account_from_dict(info, scopes=scopes)
        self.configuration = configuration
        if updatesheets is None:
            updatesheets = self.configuration.keys()
            logger.info("Updating all spreadsheets")
        else:
            logger.info(f"Updating only these spreadsheets: {updatesheets}")
        self.updatesheets = updatesheets
        self.tabs = tabs

    def update_tab(
        self,
        tabname: str,
        values: Union[List, DataFrame],
        hxltags: Optional[Dict] = None,
        limit: Optional[int] = None,
    ) -> None:
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Values in a list of lists or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.
            limit (Optional[int]): Maximum number of rows to output

        Returns:
            None
        """
        if tabname not in self.updatetabs:
            return
        for sheet in self.configuration:
            if sheet not in self.updatesheets:
                continue
            url = self.configuration[sheet]
            spreadsheet = self.gc.open_by_url(url)

            tab = spreadsheet.worksheet(self.tabs[tabname])
            tab.clear()
            if not isinstance(values, list):
                headers = list(values.columns.values)
                rows = [headers]
                if hxltags:
                    rows.append(
                        [hxltags.get(header, "") for header in headers]
                    )
                if limit is not None:
                    values = values.head(limit)
                df = values.copy(deep=True)
                df.replace(numpy.inf, "inf", inplace=True)
                df.replace(-numpy.inf, "-inf", inplace=True)
                df.fillna("NaN", inplace=True)
                rows.extend(df.values.tolist())
                values = rows
            tab.update("A1", values)



import logging
from os.path import join
from typing import Any, Dict, List, Optional, Union

from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.saver import save_json

from .base import BaseOutput

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None


from ..utilities import match_template
from ..utilities.reader import Read

logger = logging.getLogger(__name__)


class JsonFile(BaseOutput):
    """JsonFile class enabling writing to JSON files.

    Args:
        configuration (Dict): Configuration for Google Sheets
        updatetabs (List[str]): Tabs to update
        suffix (str): A suffix to add to keys. Default is _data.
    """

    def __init__(self, configuration, updatetabs, suffix="_data"):
        super().__init__(updatetabs)
        self.configuration = configuration
        self.json = dict()
        self.suffix = suffix

    def add_data_row(self, key: str, row: Dict) -> None:
        """Add row to JSON under a key

        Args:
            key (str): Key in JSON to update
            rows (List[Dict]): List of dictionaries

        Returns:
            None
        """
        dict_of_lists_add(self.json, f"{key}{self.suffix}", row)

    def add_dataframe_rows(
        self, key: str, df: DataFrame, hxltags: Optional[Dict] = None
    ) -> None:
        """Add rows from dataframe under a key

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        if hxltags:
            df = df.rename(columns=hxltags)
        self.json[f"{key}{self.suffix}"] = df.to_dict(orient="records")

    def add_data_rows_by_key(
        self,
        key: str,
        countryiso: str,
        rows: List[Dict],
        hxltags: Optional[Dict] = None,
    ) -> None:
        """Add rows under both a key and an ISO 3 country code subkey

        Args:
            key (str): Key in JSON to update
            countryiso (str): Country to use as subkey
            rows (List[Dict]): List of dictionaries
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        fullname = f"{key}{self.suffix}"
        jsondict = self.json.get(fullname, dict())
        jsondict[countryiso] = list()
        for row in rows:
            if hxltags:
                newrow = dict()
                for header, hxltag in hxltags.items():
                    newrow[hxltag] = row[header]
            else:
                newrow = row
            jsondict[countryiso].append(newrow)
        self.json[fullname] = jsondict

    def generate_json_from_list(self, key: str, rows: List[Dict]) -> None:
        """Generate JSON from key and rows list

        Args:
            key (str): Key in JSON to update
            rows (List[Dict]): List of dictionaries

        Returns:
            None
        """
        hxltags = rows[1]
        for row in rows[2:]:
            newrow = dict()
            for i, hxltag in enumerate(hxltags):
                value = row[i]
                if value in [None, ""]:
                    continue
                newrow[hxltag] = str(value)
            self.add_data_row(key, newrow)

    def generate_json_from_df(
        self, key: str, df: DataFrame, hxltags: Optional[Dict]
    ) -> None:
        """Generate JSON from key and dataframe

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        for i, row in df.iterrows():
            newrow = dict()
            row = row.to_dict()
            for i, hxltag in enumerate(hxltags):
                value = row.get(hxltag)
                if value in [None, ""]:
                    value = None
                newrow[hxltags.get(hxltag)] = str(value)
            self.add_data_row(key, newrow)

    def update_tab(
        self,
        tabname: str,
        values: Union[List, DataFrame],
        hxltags: Optional[Dict] = None,
    ) -> None:
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Values in a list of lists or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        if tabname not in self.updatetabs:
            return
        if isinstance(values, list):
            self.generate_json_from_list(tabname, values)
        else:
            # isinstance(values, DataFrame)
            self.generate_json_from_df(tabname, values, hxltags)

    def add_additional(self) -> None:
        """Download JSON files and add them under keys defined in the configuration

        Returns:
            None
        """
        reader = Read.get_reader()
        for datasetinfo in self.configuration.get("additional_inputs", list()):
            headers, iterator = reader.read(datasetinfo)
            hxl_row = next(iterator)
            if not isinstance(hxl_row, dict):
                hxl_row = hxl_row.value
            name = datasetinfo["name"]
            for row in iterator:
                newrow = dict()
                if not isinstance(row, dict):
                    row = row.value
                for key in row:
                    hxltag = hxl_row[key]
                    if hxltag != "":
                        newrow[hxl_row[key]] = row[key]
                self.add_data_row(name, newrow)

    def save(self, folder: Optional[str] = None, **kwargs: Any) -> List[str]:
        """Save JSON file and any addition subsets of that JSON defined in the additional configuration

        Args:
            folder (Optional[str]): Folder to save to. Defaults to None.
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            List[str]: List of file paths
        """
        filepaths = list()
        filepath = self.configuration["output"]
        if folder:
            filepath = join(folder, filepath)
        logger.info(f"Writing JSON to {filepath}")
        save_json(self.json, filepath)
        filepaths.append(filepath)
        for kwarg in kwargs:
            exec(f"{kwarg}={kwargs[kwarg]}")
        additional = self.configuration.get("additional_outputs", list())
        for filedetails in additional:
            json = dict()
            remove = filedetails.get("remove")
            if remove is None:
                tabs = filedetails["tabs"]
            else:
                tabs = list()
                for key in self.json.keys():
                    tab = key.replace(f"{self.suffix}", "")
                    if tab not in remove:
                        tabs.append({"tab": tab})
            for tabdetails in tabs:
                key = f'{tabdetails["tab"]}{self.suffix}'
                newjson = self.json.get(key)
                filters = tabdetails.get("filters", dict())
                hxltags = tabdetails.get("output")
                if (filters or hxltags or remove) and isinstance(
                    newjson, list
                ):
                    rows = list()
                    for row in newjson:
                        ignore_row = False
                        for filter, allowed_values in filters.items():
                            value = row.get(filter)
                            if value:
                                if isinstance(allowed_values, str):
                                    (
                                        template_string,
                                        match_string,
                                    ) = match_template(allowed_values)
                                    if template_string:
                                        allowed_values = eval(
                                            allowed_values.replace(
                                                template_string, match_string
                                            )
                                        )
                                if isinstance(allowed_values, list):
                                    if value not in allowed_values:
                                        ignore_row = True
                                        break
                                elif value != allowed_values:
                                    ignore_row = True
                                    break
                        if ignore_row:
                            continue
                        if hxltags is None:
                            newrow = row
                        else:
                            newrow = dict()
                            for hxltag in hxltags:
                                if hxltag in row:
                                    newrow[hxltag] = row[hxltag]
                        rows.append(newrow)
                    newjson = rows
                newkey = tabdetails.get("key", key)
                json[newkey] = newjson
            if not json:
                continue
            filedetailspath = filedetails["filepath"]
            if folder:
                filedetailspath = join(folder, filedetailspath)
            logger.info(f"Writing JSON to {filedetailspath}")
            save_json(json, filedetailspath)
            filepaths.append(filedetailspath)
        return filepaths



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, List, Optional, Tuple

from hdx.scraper.utilities.reader import Read


class BaseScraper(ABC):
    """Base scraper class for scrapers to inherit

    Args:
        name (str): Name of scraper
        datasetinfo (Dict): Information about dataset
        headers: Dict[str, Tuple]: Headers to be oytput at each level_name
    """

    population_lookup = dict()

    def __init__(
        self, name: str, datasetinfo: Dict, headers: Dict[str, Tuple]
    ):
        """
        Create values member variable for inheriting scrapers to populate. It is of
        form: {"national": ({"AFG": 1.2, "PSE": 1.4}, {"AFG": 123, "PSE": 241}, ...)}}
        It should be called last from the inheriting scraper's constructor.
        """
        self.setup(name, headers)
        self.datasetinfo = deepcopy(datasetinfo)
        self.errors_on_exit = None
        self.can_fallback = True

    def setup(self, name: str, headers: Dict[str, Tuple]):
        self.name = name
        self.headers = headers
        self.initialise_values_sources()
        self.has_run = False
        self.fallbacks_used = False
        self.source_urls = set()

    def get_reader(
        self, name: Optional[str] = None, prefix: Optional[str] = None
    ):
        if not name:
            name = self.name
        reader = Read.get_reader(name)
        if not prefix:
            prefix = name
        reader.prefix = prefix
        return reader

    def initialise_values_sources(self):
        self.values: Dict[str, Tuple] = {
            level: tuple(dict() for _ in value[0])
            for level, value in self.headers.items()
        }
        self.sources: Dict[str, List] = {
            level: list() for level in self.headers
        }

    def get_headers(self, level: str) -> Optional[Tuple[Tuple]]:
        """
        Get headers for a particular level_name like national or subnational

        Args:
            level (str): Level for which to get headers

        Returns:
            Optional[Tuple[Tuple]]: Scraper headers or None
        """
        return self.headers.get(level)

    def get_values(self, level: str) -> Optional[Tuple]:
        """
        Get values for a particular level_name like national or subnational

        Args:
            level (str): Level for which to get headers

        Returns:
            Optional[Tuple]: Scraper values or None
        """
        return self.values.get(level)

    def add_sources(self):
        """
        Adds sources for a particular level_name

        Returns:
            List[Tuple]: List of (hxltag, date, source, source url)

        """
        source = self.datasetinfo["source"]
        if isinstance(source, str):
            source = {"default_source": source}
        source_url = self.datasetinfo["source_url"]
        if isinstance(source_url, str):
            source_url = {"default_url": source_url}

        date = self.datasetinfo["source_date"].strftime("%Y-%m-%d")

        for level in self.headers:
            self.sources[level] = [
                (
                    hxltag,
                    date,
                    source.get(hxltag, source["default_source"]),
                    source_url.get(hxltag, source_url["default_url"]),
                )
                for hxltag in self.headers[level][1]
            ]

    def add_hxltag_source(self, tab: str, hxltag: str) -> None:
        self.sources[tab] = [
            (
                hxltag,
                self.datasetinfo["source_date"].strftime("%Y-%m-%d"),
                self.datasetinfo["source"],
                self.datasetinfo["source_url"],
            )
        ]

    def get_sources(self, level):
        return self.sources.get(level)

    def add_source_urls(self) -> None:
        source_url = self.datasetinfo.get("source_url")
        if source_url:
            if isinstance(source_url, str):
                self.source_urls.add(source_url)
            else:
                for url in source_url.values():
                    self.source_urls.add(url)

    def get_source_urls(self):
        return self.source_urls

    def add_population(self):
        for level in self.headers:
            try:
                population_index = self.headers[level][1].index("#population")
            except ValueError:
                population_index = None
            if population_index is not None:
                for key, value in self.values[level][population_index].items():
                    try:
                        valint = int(value)
                        self.population_lookup[key] = valint
                    except ValueError:
                        pass

    @abstractmethod
    def run(self) -> None:
        """
        Run scraper

        Returns:
            None
        """

    def run_after_fallbacks(self) -> None:
        """
        Executed after fallbacks are used

        Returns:
            None
        """
        pass

    def post_run(self) -> None:
        """
        Executed after running

        Returns:
            None
        """
        pass



import filecmp
from os import getenv
from os.path import join

import numpy as np
import pandas
import pytest
from hdx.utilities.path import temp_dir

from hdx.scraper.outputs.base import BaseOutput
from hdx.scraper.outputs.excelfile import ExcelFile
from hdx.scraper.outputs.googlesheets import GoogleSheets
from hdx.scraper.outputs.json import JsonFile


class TestOutput:
    @pytest.fixture(scope="class")
    def hxltags(self):
        return {
            "Country Code": "#country+code",
            "Country Name": "#country+name",
            "Population": "#population",
        }

    def test_save(self, configuration, fixtures, hxltags):
        with temp_dir(
            "TestScraperSave", delete_on_success=True, delete_on_failure=False
        ) as tempdir:
            tabs = configuration["tabs"]
            sheetname = list(tabs.values())[0]
            noout = BaseOutput(tabs)
            excelout = ExcelFile(join(tempdir, "test_output.xlsx"), tabs, tabs)
            gsheet_auth = getenv("GSHEET_AUTH")
            if not gsheet_auth:
                raise ValueError("No gsheet authorisation supplied!")
            googleout = GoogleSheets(
                configuration["googlesheets"],
                gsheet_auth,
                None,
                tabs,
                tabs,
            )
            jsonout = JsonFile(configuration["json"], tabs)
            output = [
                list(hxltags.keys()),
                list(hxltags.values()),
                ["AFG", "Afghanistan", 38041754],
            ]

            # won't do anything as wrong tab name
            excelout.update_tab("lala", output, hxltags=hxltags)
            googleout.update_tab("lala", output, hxltags=hxltags)
            jsonout.update_tab("lala", output, hxltags=hxltags)

            noout.update_tab("national", output, hxltags=hxltags)
            excelout.update_tab("national", output, hxltags=hxltags)
            googleout.update_tab("national", output, hxltags=hxltags)
            jsonout.update_tab("national", output, hxltags=hxltags)
            noout.add_additional()
            jsonout.add_additional()
            noout.save()
            excelout.save()
            filepaths = jsonout.save(tempdir, countries_to_save=["AFG"])
            excelsheet = excelout.workbook[sheetname]

            def get_list_from_cells(cells):
                result = [list(), list(), list()]
                for i, row in enumerate(excelsheet[cells]):
                    for column in row:
                        result[i].append(column.value)
                return result

            assert get_list_from_cells("A1:C3") == output
            spreadsheet = googleout.gc.open_by_url(
                configuration["googlesheets"]["test"]
            )
            googletab = spreadsheet.worksheet(sheetname)
            result = googletab.get("A1:C3")
            result[2][2] = int(result[2][2])
            assert result == output
            assert filecmp.cmp(
                filepaths[0], join(fixtures, "test_scraper_all.json")
            )
            assert filecmp.cmp(
                filepaths[1],
                join(fixtures, "test_scraper_population.json"),
            )
            assert filecmp.cmp(
                filepaths[2],
                join(fixtures, "test_scraper_population.json"),
            )
            assert filecmp.cmp(
                filepaths[3], join(fixtures, "test_scraper_other.json")
            )

            jsonout.json = dict()
            df = pandas.DataFrame(output[2:], columns=output[0])
            noout.update_tab("national", df, hxltags=hxltags)
            excelout.update_tab("national", df, hxltags=hxltags)
            googleout.update_tab("national", df, hxltags=hxltags)
            jsonout.update_tab("national", df, hxltags=hxltags)
            jsonout.add_additional()
            filepaths = jsonout.save(tempdir, countries_to_save=["AFG"])
            assert get_list_from_cells("A1:C3") == output
            result = googletab.get("A1:C3")
            result[2][2] = int(result[2][2])
            assert result == output
            assert filecmp.cmp(
                filepaths[0], join(fixtures, "test_scraper_all.json")
            )
            assert filecmp.cmp(
                filepaths[1],
                join(fixtures, "test_scraper_population.json"),
            )
            assert filecmp.cmp(
                filepaths[2],
                join(fixtures, "test_scraper_population.json"),
            )
            assert filecmp.cmp(
                filepaths[3], join(fixtures, "test_scraper_other.json")
            )

            df = pandas.DataFrame(output[1:], columns=output[0])
            googleout.update_tab("national", df, limit=2)
            result = googletab.get("A1:C3")
            result[2][2] = int(result[2][2])
            assert result == output

            output = [
                list(hxltags.keys()),
                list(hxltags.values()),
                ["AFG", "Afghanistan", 38041754],
                ["BDI", "Burundi", np.NaN],
                ["PAK", "Pakistan", -np.inf],
            ]
            df = pandas.DataFrame(output[1:], columns=output[0])
            googleout.update_tab("national", df)
            result = googletab.get("A1:C5")
            result[2][2] = int(result[2][2])
            assert result[3][2] == "NaN"
            result[3][2] = np.NaN
            assert result[4][2] == "-inf"
            result[4][2] = -np.inf
            assert result == output

    def test_jsonoutput(self, configuration, fixtures, hxltags):
        tabs = configuration["tabs"]
        noout = BaseOutput(tabs)
        jsonout = JsonFile(configuration["json"], tabs)
        rows = [
            {
                "Country Code": "#country+code",
                "Country Name": "#country+name",
                "Population": "#population",
            },
            {
                "Country Code": "AFG",
                "Country Name": "Afghanistan",
                "Population": 38041754,
            },
        ]
        noout.add_data_rows_by_key("test", "AFG", rows, hxltags)
        jsonout.add_data_rows_by_key("test", "AFG", rows, hxltags)
        assert jsonout.json == {
            "test_data": {
                "AFG": [
                    {
                        "#country+code": "#country+code",
                        "#country+name": "#country+name",
                        "#population": "#population",
                    },
                    {
                        "#country+code": "AFG",
                        "#country+name": "Afghanistan",
                        "#population": 38041754,
                    },
                ]
            }
        }
        df = pandas.DataFrame.from_records([rows[1]])
        jsonout.json = dict()
        noout.add_dataframe_rows("test", df, rows[0])
        jsonout.add_dataframe_rows("test", df, rows[0])
        assert jsonout.json == {
            "test_data": [
                {
                    "#country+code": "AFG",
                    "#country+name": "Afghanistan",
                    "#population": 38041754,
                }
            ]
        }

        noout.add_data_row("test", rows[1])  # doesn't do anything



import re
from datetime import datetime
from typing import Any, Dict, Optional, Tuple

from hdx.data.dataset import Dataset

template = re.compile("{{.*?}}")


def string_params_to_dict(string: str) -> Dict[str, str]:
    params = dict()
    if not string:
        params
    for name_par in string.split(","):
        name, par = name_par.strip().split(":")
        params[name] = par.strip()
    return params


def match_template(input: str) -> Tuple[Optional[str], Optional[str]]:
    """Try to match {{XXX}} in input string

    Args:
        input (str): String in which to look for template

    Returns:
        Tuple[Optional[str], Optional[str]]: (Matched string with brackets, matched string without brackets)
    """
    match = template.search(input)
    if match:
        template_string = match.group()
        return template_string, template_string[2:-2]
    return None, None


def get_rowval(row: Dict, valcol: str) -> Any:
    """Get the value of a particular column in a row expanding any template it contains

    Args:
        row (Dict): Dictionary
        valcol (str): Column which may be a string in which to look for template

    Returns:
        Any: Value of column or template
    """
    if "{{" in valcol:
        repvalcol = valcol
        for match in template.finditer(valcol):
            template_string = match.group()
            replace_string = f'row["{template_string[2:-2]}"]'
            repvalcol = repvalcol.replace(template_string, replace_string)
        return eval(repvalcol)
    else:
        result = row[valcol]
        if isinstance(result, str):
            return result.strip()
        return result


def get_date_from_dataset_date(
    dataset: Dataset, today: Optional[datetime] = None
) -> Optional[datetime]:
    """Return the date or end date of a dataset

    Args:
        dataset (Dataset): Dataset object
        today (Optional[datetime]): Date to use for today. Defaults to None (datetime.now())

    Returns:
        Optional[datetime]: Date or end date of a dataset
    """
    if today is None:
        date_info = dataset.get_date_of_dataset()
    else:
        date_info = dataset.get_date_of_dataset(today=today)
    return date_info.get("enddate")


def get_isodate_from_dataset_date(
    dataset: Dataset, today: Optional[datetime] = None
) -> Optional[str]:
    """Return the date or end date of a dataset as an iso formatted date

    Args:
        dataset (Dataset): Dataset object
        today (Optional[datetime]): Date to use for today. Defaults to None (datetime.now())

    Returns:
        Optional[str]: Date or end date of a dataset
    """
    date = get_date_from_dataset_date(dataset, today)
    if date:
        date = date.strftime("%Y-%m-%d")
    return date



import logging
from copy import deepcopy

from hdx.location.country import Country

logger = logging.getLogger(__name__)

regional_headers = (("regionnames",), ("#region+name",))
national_headers = (
    ["iso3", "countryname"],
    ["#country+code", "#country+name"],
)
subnational_headers = (
    ("iso3", "countryname", "adm1_pcode", "adm1_name"),
    ("#country+code", "#country+name", "#adm1+code", "#adm1+name"),
)
sources_headers = (
    ("Indicator", "Date", "Source", "Url"),
    ("#indicator+name", "#date", "#meta+source", "#meta+url"),
)


def update_tab(outputs, name, data):
    if not data:
        return
    logger.info(f"Updating tab: {name}")
    for output in outputs.values():
        output.update_tab(name, data)


def get_toplevel_rows(
    runner, names=None, overrides=dict(), toplevel="allregions"
):
    return runner.get_rows(
        toplevel, ("value",), names=names, overrides=overrides
    )


def get_regional_rows(runner, regional, names=None, level="regional"):
    return runner.get_rows(
        level, regional, regional_headers, (lambda adm: adm,), names=names
    )


def update_toplevel(
    outputs,
    toplevel_rows,
    tab="allregions",
    regional_rows=tuple(),
    regional_adm="ALL",
    regional_hxltags=None,
    regional_first=False,
):
    if not toplevel_rows:
        toplevel_rows = [list(), list(), list()]
    if regional_rows:
        adm_header = regional_rows[1].index("#region+name")
        rows_to_insert = (list(), list(), list())
        for row in regional_rows[2:]:
            if row[adm_header] == regional_adm:
                for i, hxltag in enumerate(regional_rows[1]):
                    if hxltag == "#region+name":
                        continue
                    if regional_hxltags and hxltag not in regional_hxltags:
                        continue
                    rows_to_insert[0].append(regional_rows[0][i])
                    rows_to_insert[1].append(hxltag)
                    rows_to_insert[2].append(row[i])
        if regional_first:
            toplevel_rows[0] = rows_to_insert[0] + toplevel_rows[0]
            toplevel_rows[1] = rows_to_insert[1] + toplevel_rows[1]
            toplevel_rows[2] = rows_to_insert[2] + toplevel_rows[2]
        else:
            toplevel_rows[0] += rows_to_insert[0]
            toplevel_rows[1] += rows_to_insert[1]
            toplevel_rows[2] += rows_to_insert[2]
    update_tab(outputs, tab, toplevel_rows)


def update_regional(
    outputs,
    regional_rows,
    toplevel_rows=tuple(),
    toplevel_hxltags=tuple(),
    tab="regional",
    toplevel="allregions",
):
    if not regional_rows:
        return
    toplevel_values = dict()
    toplevel_headers = dict()
    if toplevel_rows:
        for i, hxltag in enumerate(toplevel_rows[1]):
            if hxltag in toplevel_hxltags:
                toplevel_values[hxltag] = toplevel_rows[2][i]
                if hxltag not in regional_rows[1]:
                    toplevel_headers[hxltag] = toplevel_rows[0][i]
    adm_header = regional_rows[1].index("#region+name")
    found_adm = False

    def add_value(row):
        value_found = False
        for i, hxltag in enumerate(regional_rows[1]):
            value = toplevel_values.get(hxltag)
            if value is None:
                continue
            row[i] = value
            value_found = True
        for hxltag, header in toplevel_headers.items():
            value = toplevel_values.get(hxltag)
            if value is None:
                continue
            regional_rows[0].append(header)
            regional_rows[1].append(hxltag)
            row.append(value)
            value_found = True
        return value_found

    for row in regional_rows[2:]:
        if row[adm_header] == toplevel:
            add_value(row)
            found_adm = True
            break
    if not found_adm:
        row = [toplevel]
        for _ in regional_rows[0][1:]:
            row.append(None)
        if add_value(row):
            regional_rows.append(row)
    length = len(regional_rows[0])
    for row in regional_rows[2:]:
        while len(row) < length:
            row.append(None)
    update_tab(outputs, tab, regional_rows)


def update_national(
    runner,
    countries,
    outputs,
    names=None,
    flag_countries=None,
    iso3_to_region=None,
    ignore_regions=tuple(),
    level="national",
    tab="national",
):
    headers = deepcopy(national_headers)
    fns = [
        lambda adm: adm,
        lambda adm: Country.get_country_name_from_iso3(adm),
    ]

    if flag_countries:
        headers[0].append(flag_countries["header"])
        headers[1].append(flag_countries["hxltag"])
        isfc_fn = (
            lambda adm: "Y" if adm in flag_countries["countries"] else "N"
        )
        fns.append(isfc_fn)

    if iso3_to_region:
        headers[0].append("region")
        headers[1].append("#region+name")

        def region_fn(adm):
            regions = sorted(list(iso3_to_region[adm]))
            for region in reversed(regions):
                if region in ignore_regions:
                    regions.remove(region)
            return "|".join(regions)

        fns.append(region_fn)

    rows = runner.get_rows(level, countries, headers, fns, names=names)
    if rows:
        update_tab(outputs, tab, rows)


def update_subnational(
    runner,
    adminone,
    outputs,
    names=None,
    level="subnational",
    tab="subnational",
):
    def get_country_name(adm):
        countryiso3 = adminone.pcode_to_iso3[adm]
        return Country.get_country_name_from_iso3(countryiso3)

    fns = (
        lambda adm: adminone.pcode_to_iso3[adm],
        get_country_name,
        lambda adm: adm,
        lambda adm: adminone.pcode_to_name[adm],
    )
    rows = runner.get_rows(
        level, adminone.pcodes, subnational_headers, fns, names=names
    )
    update_tab(outputs, tab, rows)


def update_sources(
    runner,
    outputs,
    additional_sources=tuple(),
    names=None,
    secondary_runner=None,
    custom_sources=list(),
    tab="sources",
):
    sources = runner.get_sources(
        names=names,
        additional_sources=additional_sources,
    )
    if secondary_runner:
        secondary_sources = secondary_runner.get_sources()
        sources.extend(secondary_sources)
    sources.extend(custom_sources)
    update_tab(outputs, tab, list(sources_headers) + sources)



import logging
from datetime import datetime

from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class TimeSeries(BaseScraper):
    def __init__(self, name, datasetinfo, today, outputs):
        # Time series only outputs to separate tabs
        super().__init__(f"timeseries_{name}", datasetinfo, dict())
        self.today = today
        self.outputs = outputs

    def run(self):
        input = self.datasetinfo["input"]
        datecol = self.datasetinfo["date"]
        datetype = self.datasetinfo["date_type"]
        ignore_future_date = self.datasetinfo.get("ignore_future_date", True)
        headers = [datecol] + self.datasetinfo["output"]
        hxltags = [self.datasetinfo["date_hxl"]] + self.datasetinfo[
            "output_hxl"
        ]
        rows = [headers, hxltags]
        file_headers, iterator = self.get_reader().read(self.datasetinfo)
        for inrow in iterator:
            if isinstance(datecol, list):
                dates = [str(inrow[x]) for x in datecol]
                date = "".join(dates)
            else:
                date = inrow[datecol]
            if datetype == "date":
                if not isinstance(date, datetime):
                    date = parse_date(date)
                date = date.replace(tzinfo=None)
                if date > self.today and ignore_future_date:
                    continue
                date = date.strftime("%Y-%m-%d")
            elif datetype == "year":
                date = int(date)
                if date > self.today.year and ignore_future_date:
                    continue
                date = str(date)
            row = [date]
            for column in input:
                row.append(inrow[column])
            rows.append(row)
        for output in self.outputs.values():
            output.update_tab(self.name, rows)

    def add_sources(self):
        for hxltag in self.datasetinfo["output_hxl"]:
            self.add_hxltag_source(self.name, hxltag)

    @classmethod
    def get_scrapers(cls, configuration, today, outputs):
        scrapers = list()
        for name, datasetinfo in configuration.items():
            scrapers.append(cls(name, datasetinfo, today, outputs))
        return scrapers



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date
from hdx.utilities.errors_onexit import ErrorsOnExit

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.outputs.update_tabs import update_national
from hdx.scraper.runner import Runner
from hdx.scraper.utilities.reader import Read

from .conftest import run_check_scraper, run_check_scrapers
from .unhcr_myanmar_idps import idps_post_run


class TestScrapersNational:
    def test_get_national_afg(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG",)
        runner = Runner(iso3s, adminone, today)
        keys = runner.add_configurables(scraper_configuration, level)
        assert keys == [
            "population",
            "who_national",
            "who_national2",
            "access",
            "sadd",
            "ourworldindata",
            "broken_owd_url",
            "covidtests",
            "idps",
            "casualties",
        ]

        def assert_auth_header(name, expected):
            downloader = Read.get_reader(name).downloader
            assert downloader.session.headers.get("Authorization") == expected

        assert_auth_header("population", "pop_12345")
        assert_auth_header("who_national", "who_abc")

        def assert_auth(name, expected):
            downloader = Read.get_reader(name).downloader
            assert downloader.session.auth == expected

        assert_auth("access", ("acc_12345", "acc_abc"))
        assert_auth("who_national2", ("who_def", "who_12345"))

        def assert_params(name, expected):
            downloader = Read.get_reader(name).downloader
            assert downloader.session.params == expected

        assert_params("sadd", {"user": "sadd_123", "pass": "sadd_abc"})
        assert_params("ourworldindata", {"auth": "owid_abc"})

        name = "population"
        headers = (["Population"], ["#population"])
        values = [{"AFG": 38041754}]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            population_lookup=values[0],
            source_urls=[
                "https://data.humdata.org/organization/world-bank-group"
            ],
        )

        names = ("who_national", "who_national2")
        headers = (
            [
                "CasesPer100000",
                "DeathsPer100000",
                "Cases2Per100000",
                "Deaths2Per100000",
            ],
            [
                "#affected+infected+per100000",
                "#affected+killed+per100000",
                "#affected+infected+2+per100000",
                "#affected+killed+2+per100000",
            ],
        )
        values = [
            {"AFG": "96.99"},
            {"AFG": "3.41"},
            {"AFG": "96.99"},
            {"AFG": "3.41"},
        ]
        sources = [
            (
                "#affected+infected+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+infected+2+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+2+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://covid19.who.int/WHO-COVID-19-global-data.csv"
            ],
            set_not_run=False,
        )

        def passthrough_fn(x):
            return x

        fns = (passthrough_fn,)
        rows = runner.get_rows(
            "national", iso3s, (("iso3",), ("#country+code",)), fns, names
        )
        assert rows == [
            [
                "iso3",
                "CasesPer100000",
                "DeathsPer100000",
                "Cases2Per100000",
                "Deaths2Per100000",
            ],
            [
                "#country+code",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
                "#affected+infected+2+per100000",
                "#affected+killed+2+per100000",
            ],
            ["AFG", "96.99", "3.41", "96.99", "3.41"],
        ]
        combined_sources = [
            (
                "#date+start+conflict",
                "2022-02-24",
                "Meduza",
                "https://meduza.io/en/news/2022/02/24/putin-announces-start-of-military-operation-in-eastern-ukraine",
            ),
        ] + sources
        assert (
            runner.get_sources(
                additional_sources=configuration["additional_sources"]
            )
            == combined_sources
        )
        runner.set_not_run_many(names)

        name = "access"
        headers = (
            [
                "% of visas pending or denied",
                "% of travel authorizations or movements denied",
                "Number of incidents reported in previous year",
                "Number of incidents reported since start of year",
                "Number of incidents reported since start of previous year",
                "% of CERF projects affected by insecurity and inaccessibility",
                "% of CBPF projects affected by insecurity and inaccessibility",
                "Campaign Vaccine",
                "Campaign Vaccine Status",
                "Number of learners enrolled from pre-primary to tertiary education",
            ],
            [
                "#access+visas+pct",
                "#access+travel+pct",
                "#event+year+previous+num",
                "#event+year+todate+num",
                "#event+year+previous+todate+num",
                "#activity+cerf+project+insecurity+pct",
                "#activity+cbpf+project+insecurity+pct",
                "#service+name",
                "#status+name",
                "#population+education",
            ],
        )
        values = [
            {"AFG": 0.2},
            {"AFG": "N/A"},
            {"AFG": "20"},
            {"AFG": "2"},
            {"AFG": "22"},
            {"AFG": 0.5710000000000001},
            {"AFG": 0.04},
            {"AFG": "bivalent Oral Poliovirus"},
            {"AFG": "Postponed"},
            {"AFG": 9979405},
        ]
        sources = [
            (
                "#access+visas+pct",
                "2020-10-01",
                "OCHA",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#access+travel+pct",
                "2020-10-01",
                "OCHA",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#event+year+previous+num",
                "2020-10-01",
                "Aid Workers Database",
                "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
            ),
            (
                "#event+year+todate+num",
                "2020-10-01",
                "Aid Workers Database",
                "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
            ),
            (
                "#event+year+previous+todate+num",
                "2020-10-01",
                "Aid Workers Database",
                "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
            ),
            (
                "#activity+cerf+project+insecurity+pct",
                "2020-10-01",
                "UNCERF",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#activity+cbpf+project+insecurity+pct",
                "2020-10-01",
                "UNCERF",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#service+name",
                "2020-10-01",
                "Multiple sources",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#status+name",
                "2020-10-01",
                "Multiple sources",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#population+education",
                "2020-10-01",
                "UNESCO",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ],
        )

        name = "sadd"
        headers = (
            [
                "Cases (% male)",
                "Cases (% female)",
                "Deaths (% male)",
                "Deaths (% female)",
            ],
            [
                "#affected+infected+m+pct",
                "#affected+f+infected+pct",
                "#affected+killed+m+pct",
                "#affected+f+killed+pct",
            ],
        )
        values = [
            {"AFG": "0.7044"},
            {"AFG": "0.2956"},
            {"AFG": "0.7498"},
            {"AFG": "0.2502"},
        ]
        sources = [
            (
                "#affected+infected+m+pct",
                "2020-08-07",
                "SADD",
                "https://globalhealth5050.org/?_covid-data=dataset-fullvars&_extype=csv",
            ),
            (
                "#affected+f+infected+pct",
                "2020-08-07",
                "SADD",
                "https://globalhealth5050.org/?_covid-data=dataset-fullvars&_extype=csv",
            ),
            (
                "#affected+killed+m+pct",
                "2020-08-07",
                "SADD",
                "https://globalhealth5050.org/?_covid-data=dataset-fullvars&_extype=csv",
            ),
            (
                "#affected+f+killed+pct",
                "2020-08-07",
                "SADD",
                "https://globalhealth5050.org/?_covid-data=dataset-fullvars&_extype=csv",
            ),
        ]
        run_check_scraper(name, runner, level, headers, values, sources)

    def test_get_national_ukr(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        runner = Runner(("UKR",), adminone, today)
        runner.add_configurables(scraper_configuration, level)

        name = "casualties"
        headers = (
            ["CiviliansKilled", "CiviliansInjured"],
            ["#affected+killed", "#affected+injured"],
        )
        values = [{"UKR": "351"}, {"UKR": "707"}]
        sources = [
            (
                "#affected+killed",
                "2020-10-01",
                "OHCHR",
                "https://data.humdata.org/dataset/ukraine-key-figures-2022",
            ),
            (
                "#affected+injured",
                "2020-10-01",
                "OHCHR",
                "https://data.humdata.org/dataset/ukraine-key-figures-2022",
            ),
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/ukraine-key-figures-2022"
            ],
        )

    def test_get_national_afg_phl(self, configuration, fallbacks):
        BaseScraper.population_lookup = {"AFG": 38041754}
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        runner = Runner(("AFG", "PHL"), adminone, today)
        runner.add_configurables(scraper_configuration, level)

        name = "ourworldindata"
        headers = (
            [
                "TotalDosesAdministered",
                "PopulationCoverageAdministeredDoses",
            ],
            [
                "#capacity+doses+administered+total",
                "#capacity+doses+administered+coverage+pct",
            ],
        )
        values = [dict(), dict()]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2020-10-01",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
            (
                "#capacity+doses+administered+coverage+pct",
                "2020-10-01",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
        ]
        run_check_scraper(name, runner, level, headers, values, sources)

        name = "covidtests"
        headers = (
            [
                "New Tests",
                "New Tests Per Thousand",
                "New Tests Per Thousand (7-day)",
                "Positive Test Rate",
            ],
            [
                "#affected+tested",
                "#affected+tested+per1000",
                "#affected+tested+avg+per1000",
                "#affected+tested+positive+pct",
            ],
        )
        values = [
            {"PHL": 39611},
            {"PHL": 0.361},
            {"PHL": 0.312},
            {"PHL": 0.072},
        ]
        sources = [
            (
                "#affected+tested",
                "2020-10-01",
                "Our World in Data",
                "https://data.humdata.org/dataset/total-covid-19-tests-performed-by-country",
            ),
            (
                "#affected+tested+per1000",
                "2020-10-01",
                "Our World in Data",
                "https://data.humdata.org/dataset/total-covid-19-tests-performed-by-country",
            ),
            (
                "#affected+tested+avg+per1000",
                "2020-10-01",
                "Our World in Data",
                "https://data.humdata.org/dataset/total-covid-19-tests-performed-by-country",
            ),
            (
                "#affected+tested+positive+pct",
                "2020-10-01",
                "Our World in Data",
                "https://data.humdata.org/dataset/total-covid-19-tests-performed-by-country",
            ),
        ]
        run_check_scraper(name, runner, level, headers, values, sources)

        today = parse_date("2021-05-03")
        errors_on_exit = ErrorsOnExit()
        runner = Runner(
            ("AFG", "PHL"),
            adminone,
            today,
            errors_on_exit=errors_on_exit,
        )
        runner.add_configurables(scraper_configuration, level)
        name = "ourworldindata"
        headers = (
            [
                "TotalDosesAdministered",
                "PopulationCoverageAdministeredDoses",
            ],
            [
                "#capacity+doses+administered+total",
                "#capacity+doses+administered+coverage+pct",
            ],
        )
        values = [{"AFG": 240000}, {"AFG": "0.0032"}]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
            (
                "#capacity+doses+administered+coverage+pct",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
        ]
        run_check_scraper(name, runner, level, headers, values, sources)

        # Test fallbacks
        name = "broken_owd_url"
        headers = (
            ["TotalDosesAdministered"],
            ["#capacity+doses+administered+total"],
        )
        values = [{"AFG": "230000"}]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2020-09-01",
                "Our World in Data",
                "tests/fixtures/fallbacks.json",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            fallbacks_used=True,
        )
        assert errors_on_exit.errors == [
            "Using fallbacks for broken_owd_url! Error: [scheme-error] The data source could not be successfully loaded: [Errno 2] No such file or directory: 'tests/fixtures/broken_owd_url_notexist.csv'"
        ]

    def test_get_national_afg_mmr_phl(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2021-05-03")
        errors_on_exit = ErrorsOnExit()
        adminone = AdminOne(configuration)
        level = "national"
        countries = ("AFG", "MMR", "PHL")
        scraper_configuration = configuration[f"scraper_{level}"]

        runner = Runner(
            countries,
            adminone,
            today,
            errors_on_exit=errors_on_exit,
        )
        scrapers = runner.add_configurables(scraper_configuration, level)
        name = "idps"
        headers = (
            ["TotalIDPs"],
            ["#affected+displaced"],
        )
        values = [{"AFG": 4664000, "MMR": 509600, "PHL": 298000}]
        sources = [
            (
                "#affected+displaced",
                "2020-12-31",
                "IDMC",
                "https://data.humdata.org/dataset/idmc-internally-displaced-persons-idps",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/idmc-internally-displaced-persons-idps"
            ],
        )

        runner.add_instance_variables(
            "idps", overrideinfo=configuration["unhcr_myanmar_idps"]
        )
        runner.add_post_run("idps", idps_post_run)
        values = [{"AFG": 4664000, "MMR": 569591, "PHL": 298000}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/idmc-internally-displaced-persons-idps",
                "https://data.unhcr.org/population/?widget_id=264111&geo_id=693&population_group=5407,4999",
            ],
        )
        assert errors_on_exit.errors == list()
        runner = Runner(
            ("AFG", "MMR", "PHL"),
            adminone,
            today,
            errors_on_exit=errors_on_exit,
        )
        runner.add_configurables(scraper_configuration, level)
        runner.add_instance_variables(
            "idps", overrideinfo={"url": "NOT EXIST"}
        )
        runner.add_post_run("idps", idps_post_run)
        values = [{"AFG": 4664000, "MMR": 509600, "PHL": 298000}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/idmc-internally-displaced-persons-idps"
            ],
        )
        assert errors_on_exit.errors == [
            "Not using UNHCR Myanmar IDPs override! Error: [Errno 2] No such file or directory: 'tests/fixtures/idps_override_not-exist.json'",
        ]
        runner.run()
        jsonout = JsonFile(configuration["json"], [level])
        outputs = {"json": jsonout}
        flag_countries = {
            "header": "ishrp",
            "hxltag": "#meta+ishrp",
            "countries": ("AFG", "MMR"),
        }
        iso3_to_region = {
            "AFG": ("Region1", "Region2"),
            "PHL": ("Region3", "Region6"),
            "MMR": ("Region3", "Region4"),
        }
        update_national(
            runner,
            countries,
            outputs,
            names=scrapers,
            flag_countries=flag_countries,
            iso3_to_region=iso3_to_region,
            ignore_regions=("Region6",),
        )
        assert jsonout.json[f"{level}_data"] == [
            {
                "#access+travel+pct": "N/A",
                "#access+visas+pct": "0.2",
                "#activity+cbpf+project+insecurity+pct": "0.04",
                "#activity+cerf+project+insecurity+pct": "0.5710000000000001",
                "#affected+displaced": "4664000",
                "#affected+f+infected+pct": "0.2956",
                "#affected+f+killed+pct": "0.2502",
                "#affected+infected+2+per100000": "96.99",
                "#affected+infected+m+pct": "0.7044",
                "#affected+infected+per100000": "96.99",
                "#affected+killed+2+per100000": "3.41",
                "#affected+killed+m+pct": "0.7498",
                "#affected+killed+per100000": "3.41",
                "#capacity+doses+administered+coverage+pct": "0.0032",
                "#capacity+doses+administered+total": "230000",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#event+year+previous+num": "20",
                "#event+year+previous+todate+num": "22",
                "#event+year+todate+num": "2",
                "#meta+ishrp": "Y",
                "#population": "38041754",
                "#population+education": "9979405",
                "#region+name": "Region1|Region2",
                "#service+name": "bivalent Oral Poliovirus",
                "#status+name": "Postponed",
            },
            {
                "#affected+displaced": "509600",
                "#capacity+doses+administered+coverage+pct": "0.0096",
                "#capacity+doses+administered+total": "1040000",
                "#country+code": "MMR",
                "#country+name": "Myanmar",
                "#meta+ishrp": "Y",
                "#population": "54045420",
                "#region+name": "Region3|Region4",
            },
            {
                "#affected+displaced": "298000",
                "#affected+tested": "39407",
                "#affected+tested+avg+per1000": "0.469",
                "#affected+tested+per1000": "0.36",
                "#affected+tested+positive+pct": "0.155",
                "#country+code": "PHL",
                "#country+name": "Philippines",
                "#meta+ishrp": "N",
                "#population": "108116615",
                "#region+name": "Region3",
            },
        ]



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.outputs.update_tabs import update_subnational
from hdx.scraper.runner import Runner

from .conftest import run_check_scraper


class TestScrapersSubnational:
    def test_get_subnational(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "subnational"
        scraper_configuration = configuration[f"scraper_{level}"]
        runner = Runner(("AFG",), adminone, today)
        keys = runner.add_configurables(scraper_configuration, level)
        assert keys == ["gam"]

        name = "gam"
        headers = (
            ["Malnutrition Estimate"],
            ["#severity+malnutrition+num+subnational"],
        )
        values = [
            {
                "AF17": 3.371688,
                "AF31": 3.519166,
                "AF09": 1.524646,
                "AF21": 1.319626,
                "AF10": 1.40426,
                "AF24": 1.043487,
                "AF33": 2.745447,
                "AF29": 2.478977,
                "AF11": 1.022871,
                "AF23": 1.340286,
                "AF30": 1.677612,
                "AF32": 1.687488,
                "AF28": 0.6210205,
                "AF01": 1.282291,
                "AF27": 1.378641,
                "AF02": 3.552082,
                "AF14": 0.7653555,
                "AF15": 0.953823,
                "AF19": 1.684882,
                "AF07": 2.090165,
                "AF05": 0.9474334,
                "AF06": 2.162038,
                "AF34": 1.6455,
                "AF16": 1.927783,
                "AF12": 4.028857,
                "AF13": 9.150105,
                "AF08": 1.64338,
                "AF03": 2.742952,
                "AF20": 1.382376,
                "AF22": 1.523334,
                "AF18": 0.9578965,
                "AF25": 0.580423,
                "AF04": 0.501081,
                "AF26": 4.572629,
            }
        ]
        sources = [
            (
                "#severity+malnutrition+num+subnational",
                "2020-10-01",
                "UNICEF",
                "https://data.humdata.org/dataset/87b86e7d-e9b2-4922-a48e-1f10afd528e6/resource/eabba7e7-16d0-436c-a62c-df6edd03be7c/download/unicef_who_wb_global_expanded_databases_severe_wasting.xlsx",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            set_not_run=False,
        )

        jsonout = JsonFile(configuration["json"], [level])
        outputs = {"json": jsonout}
        update_subnational(
            runner,
            adminone,
            outputs,
            names=(name,),
        )
        assert jsonout.json[f"{level}_data"] == [
            {
                "#adm1+code": "AF01",
                "#adm1+name": "Kabul",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.282291",
            },
            {
                "#adm1+code": "AF02",
                "#adm1+name": "Kapisa",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "3.552082",
            },
            {
                "#adm1+code": "AF03",
                "#adm1+name": "Parwan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "2.742952",
            },
            {
                "#adm1+code": "AF04",
                "#adm1+name": "Maidan Wardak",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.501081",
            },
            {
                "#adm1+code": "AF05",
                "#adm1+name": "Logar",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.9474334",
            },
            {
                "#adm1+code": "AF06",
                "#adm1+name": "Nangarhar",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "2.162038",
            },
            {
                "#adm1+code": "AF07",
                "#adm1+name": "Laghman",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "2.090165",
            },
            {
                "#adm1+code": "AF08",
                "#adm1+name": "Panjsher",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.64338",
            },
            {
                "#adm1+code": "AF09",
                "#adm1+name": "Baghlan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.524646",
            },
            {
                "#adm1+code": "AF10",
                "#adm1+name": "Bamyan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.40426",
            },
            {
                "#adm1+code": "AF11",
                "#adm1+name": "Ghazni",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.022871",
            },
            {
                "#adm1+code": "AF12",
                "#adm1+name": "Paktika",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "4.028857",
            },
            {
                "#adm1+code": "AF13",
                "#adm1+name": "Paktya",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "9.150105",
            },
            {
                "#adm1+code": "AF14",
                "#adm1+name": "Khost",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.7653555",
            },
            {
                "#adm1+code": "AF15",
                "#adm1+name": "Kunar",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.953823",
            },
            {
                "#adm1+code": "AF16",
                "#adm1+name": "Nuristan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.927783",
            },
            {
                "#adm1+code": "AF17",
                "#adm1+name": "Badakhshan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "3.371688",
            },
            {
                "#adm1+code": "AF18",
                "#adm1+name": "Takhar",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.9578965",
            },
            {
                "#adm1+code": "AF19",
                "#adm1+name": "Kunduz",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.684882",
            },
            {
                "#adm1+code": "AF20",
                "#adm1+name": "Samangan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.382376",
            },
            {
                "#adm1+code": "AF21",
                "#adm1+name": "Balkh",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.319626",
            },
            {
                "#adm1+code": "AF22",
                "#adm1+name": "Sar E Pul",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.523334",
            },
            {
                "#adm1+code": "AF23",
                "#adm1+name": "Ghor",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.340286",
            },
            {
                "#adm1+code": "AF24",
                "#adm1+name": "Daykundi",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.043487",
            },
            {
                "#adm1+code": "AF25",
                "#adm1+name": "Uruzgan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.580423",
            },
            {
                "#adm1+code": "AF26",
                "#adm1+name": "Zabul",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "4.572629",
            },
            {
                "#adm1+code": "AF27",
                "#adm1+name": "Kandahar",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.378641",
            },
            {
                "#adm1+code": "AF28",
                "#adm1+name": "Jawzjan",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "0.6210205",
            },
            {
                "#adm1+code": "AF29",
                "#adm1+name": "Faryab",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "2.478977",
            },
            {
                "#adm1+code": "AF30",
                "#adm1+name": "Hilmand",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.677612",
            },
            {
                "#adm1+code": "AF31",
                "#adm1+name": "Badghis",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "3.519166",
            },
            {
                "#adm1+code": "AF32",
                "#adm1+name": "Hirat",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.687488",
            },
            {
                "#adm1+code": "AF33",
                "#adm1+name": "Farah",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "2.745447",
            },
            {
                "#adm1+code": "AF34",
                "#adm1+name": "Nimroz",
                "#country+code": "AFG",
                "#country+name": "Afghanistan",
                "#severity+malnutrition+num+subnational": "1.6455",
            },
        ]

        scraper_configuration = configuration["other"]
        runner.add_configurables(scraper_configuration, level)
        name = "gam_other"
        headers = (
            ["Malnutrition Estimate"],
            ["#severity+malnutrition+num+subnational"],
        )
        values = [{"AF09": 1.524646, "AF24": 1.043487}]
        sources = [
            (
                "#severity+malnutrition+num+subnational",
                "2020-10-01",
                "UNICEF",
                "https://data.humdata.org/dataset/87b86e7d-e9b2-4922-a48e-1f10afd528e6/resource/eabba7e7-16d0-436c-a62c-df6edd03be7c/download/unicef_who_wb_global_expanded_databases_severe_wasting.xlsx",
            )
        ]
        run_check_scraper(name, runner, level, headers, values, sources)



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.aggregator import Aggregator
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.outputs.update_tabs import (
    get_regional_rows,
    get_toplevel_rows,
    update_regional,
    update_toplevel,
)
from hdx.scraper.runner import Runner

from .conftest import run_check_scrapers


class TestScrapersRegionalToplevel:
    def test_regionaltoplevel(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "MMR")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        level = "national"
        BaseScraper.population_lookup = dict()
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        names = ("population", "who_national")
        headers = (
            [
                "Population",
                "CasesPer100000",
                "DeathsPer100000",
            ],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
            ],
        )
        national_values = [
            {"AFG": 38041754, "PSE": 4685306},
            {"AFG": "96.99", "PSE": "362.43"},
            {"AFG": "3.41", "PSE": "1.98"},
        ]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        source_urls = [
            "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            "https://data.humdata.org/organization/world-bank-group",
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            national_values,
            sources,
            source_urls=source_urls,
            set_not_run=False,
        )

        level = "global"
        configuration_hxl = configuration["aggregation_hxl"]
        aggregator_configuration = configuration_hxl[f"aggregation_{level}"]
        adm_aggregation = ("AFG", "PSE")
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        names = runner.add_customs(scrapers, add_to_run=True)
        assert names == [
            f"population_{level}",
            f"affected_infected_per100000_{level}",
            f"affected_infected_perpop_{level}",
        ]
        pop = 42727060
        headers = (
            ["Population", "CasesPer100000", "CasesPerPopulation"],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+infected+perpop",
            ],
        )
        values = [{"value": pop}, {"value": "229.71"}, {"value": "0.5376"}]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | {"global": pop},
            source_urls=source_urls,
            set_not_run=False,
        )

        regions = ("ROAP",)
        aggregator_configuration = configuration_hxl["aggregation_sum"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            "regional",
            {"AFG": regions, "PSE": ("somewhere",)},
            runner,
        )
        regional_names = runner.add_customs(scrapers, add_to_run=True)
        runner.run(regional_names)
        regional_rows = get_regional_rows(
            runner, regions, names=regional_names
        )
        toplevel = "allregions"
        mapping = {"global": toplevel}
        allregions_rows = get_toplevel_rows(
            runner,
            names=names,
            overrides={
                names[0]: mapping,
                names[1]: mapping,
                names[2]: mapping,
            },
            toplevel=toplevel,
        )
        level = "regional"
        jsonout = JsonFile(configuration["json"], [level, toplevel])
        outputs = {"json": jsonout}
        update_regional(
            outputs,
            regional_rows,
            allregions_rows,
            toplevel_hxltags=allregions_rows[1],
            tab=level,
            toplevel=toplevel,
        )
        assert jsonout.json[f"{level}_data"] == [
            {"#population": "38041754", "#region+name": "ROAP"},
            {
                "#region+name": toplevel,
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
            },
        ]
        regional_rows = (
            ("regionnames", "MyColumn"),
            ("#region+name", "#mycolumn"),
            (toplevel, 12345),
        )
        update_toplevel(
            outputs,
            allregions_rows,
            regional_rows=regional_rows,
            regional_adm=toplevel,
            regional_hxltags=["#mycolumn"],
        )
        assert jsonout.json["allregions_data"] == [
            {
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
                "#mycolumn": "12345",
            }
        ]
        jsonout.json["allregions_data"] = list()
        update_toplevel(
            outputs,
            allregions_rows,
            regional_rows=regional_rows,
            regional_adm=toplevel,
            regional_hxltags=["#mycolumn"],
            regional_first=True,
        )
        assert jsonout.json["allregions_data"] == [
            {
                "#mycolumn": "12345",
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
            }
        ]



import logging

from hdx.utilities.text import get_fraction_str

from hdx.scraper.base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class EducationEnrolment(BaseScraper):
    def __init__(self, datasetinfo, closures, countryiso3s, regionlookup):
        super().__init__(
            "education_enrolment",
            datasetinfo,
            {
                "national": (
                    (
                        "No. pre-primary to upper-secondary learners",
                        "No. tertiary learners",
                        "No. affected learners",
                    ),
                    (
                        "#population+learners+pre_primary_to_secondary",
                        "#population+learners+tertiary",
                        "#affected+learners",
                    ),
                ),
                "regional": (
                    (
                        "No. affected learners",
                        "Percentage affected learners",
                    ),
                    (
                        "#affected+learners",
                        "#affected+learners+pct",
                    ),
                ),
            },
        )
        self.closures = closures
        self.countryiso3s = countryiso3s
        self.regionlookup = regionlookup

    def run(self) -> None:
        learners_headers, learners_iterator = self.get_reader().read(
            self.datasetinfo,
        )
        learners_012, learners_3, affected_learners = self.get_values(
            "national"
        )
        all_learners = dict()

        for row in learners_iterator:
            countryiso = row["ISO3"]
            if not countryiso or countryiso not in self.countryiso3s:
                continue
            l_0 = row["Pre-primary (both)"]
            l_1 = row["Primary (both)"]
            l_2 = row["Secondary (both)"]
            l_3 = row["Tertiary (both)"]
            l_012 = None
            if l_0 is not None and l_0 != "-":
                l_012 = int(l_0)
            if l_1 is not None and l_1 != "-":
                l_1 = int(l_1)
                if l_012 is None:
                    l_012 = l_1
                else:
                    l_012 += l_1
            if l_2 is not None and l_2 != "-":
                l_2 = int(l_2)
                if l_012 is None:
                    l_012 = l_2
                else:
                    l_012 += l_2
            if l_012 is not None:
                learners_012[countryiso] = l_012
            if l_3 == "-":
                l_3 = None
            elif l_3 is not None:
                l_3 = int(l_3)
                learners_3[countryiso] = l_3
            no_learners = None
            if l_012 is not None:
                no_learners = l_012
                if l_3:
                    no_learners += l_3
            elif l_3 is not None:
                no_learners = l_3
            if no_learners is not None:
                all_learners[countryiso] = no_learners
                if countryiso in self.closures.fully_closed:
                    affected_learners[countryiso] = no_learners
        affected_learners_total = self.get_values("regional")[0]
        learners_total = dict()
        for countryiso in all_learners:
            country_learners = all_learners[countryiso]
            country_affected_learners = affected_learners.get(countryiso)
            for region in self.regionlookup.iso3_to_region_and_hrp[countryiso]:
                learners_total[region] = (
                    learners_total.get(region, 0) + country_learners
                )
                if country_affected_learners is not None:
                    affected_learners_total[region] = (
                        affected_learners_total.get(region, 0)
                        + country_affected_learners
                    )
        percentage_affected_learners = self.get_values("regional")[1]
        for region, no_learners in affected_learners_total.items():
            percentage_affected_learners[region] = get_fraction_str(
                no_learners, learners_total[region]
            )



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.timeseries import TimeSeries
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.runner import Runner


class TestScrapersTimeSeries:
    def test_timeseries(self, configuration):
        BaseScraper.population_lookup = dict()
        adminone = AdminOne(configuration)
        iso3s = ("AFG", "MMR")
        today = parse_date("2022-03-06")
        runner = Runner(iso3s, adminone, today)
        name = "timeseries_casualties"
        jsonout = JsonFile(configuration["json"], [name])
        outputs = {"json": jsonout}
        scrapers = TimeSeries.get_scrapers(
            configuration["timeseries"], today, outputs
        )
        runner.add_customs(scrapers)
        scraper_names = runner.add_customs(scrapers, add_to_run=True)
        assert scraper_names == [name]
        runner.run()
        assert jsonout.json[f"{name}_data"] == [
            {
                "#affected+injured": "553",
                "#affected+killed": "249",
                "#date": "2022-03-03",
            },
            {
                "#affected+injured": "675",
                "#affected+killed": "331",
                "#date": "2022-03-04",
            },
            {
                "#affected+injured": "707",
                "#affected+killed": "351",
                "#date": "2022-03-05",
            },
        ]



import logging
from os.path import join
from shutil import copy2

from slugify import slugify

from hdx.scraper.base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class FileCopier(BaseScraper):
    def __init__(self, datasetinfo, folder):
        # FileCopier only outputs to sources
        name = f"filecopier_{slugify(datasetinfo['hxltag'].lower(), separator='_')}"
        super().__init__(name, datasetinfo, dict())
        self.folder = folder

    def run(self):
        reader = self.get_reader("hdx")
        resource = reader.read_hdx_metadata(self.datasetinfo)
        url, path = reader.read_resource(self.name, resource)
        logger.info(f"Downloading {url} to {path}")
        copy2(path, join(self.folder, self.datasetinfo["filename"]))

    def add_sources(self):
        self.add_hxltag_source("FileCopier", self.datasetinfo["hxltag"])

    @classmethod
    def get_scrapers(cls, configuration, folder=""):
        scrapers = list()
        for datasetinfo in configuration:
            scrapers.append(cls(datasetinfo, folder))
        return scrapers



import logging
from typing import Dict

from hdx.utilities.dateparse import default_date, parse_date

from hdx.scraper.base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class EducationClosures(BaseScraper):
    def __init__(self, datasetinfo: Dict, today, countryiso3s, regionlookup):
        super().__init__(
            "education_closures",
            datasetinfo,
            {
                "national": (("School Closure",), ("#impact+type",)),
                "regional": (
                    ("No. closed countries",),
                    ("#status+country+closed",),
                ),
            },
        )
        self.today = today
        self.countryiso3s = countryiso3s
        self.regionlookup = regionlookup
        self.fully_closed = None

    @staticmethod
    def get_fully_closed(closures):
        fully_closed = list()
        if not closures:
            return fully_closed
        for countryiso, closure in closures.items():
            if closure.lower() == "closed due to covid-19":
                fully_closed.append(countryiso)
        return fully_closed

    def run(self) -> None:
        closures_headers, closures_iterator = self.get_reader().read(
            self.datasetinfo,
        )
        closures = self.get_values("national")[0]
        closed_countries = self.get_values("regional")[0]
        country_dates = dict()
        for row in closures_iterator:
            countryiso = row["ISO"]
            if not countryiso or countryiso not in self.countryiso3s:
                continue
            date = row["Date"]
            if isinstance(date, str):
                date = parse_date(date)
            if date > self.today:
                continue
            max_date = country_dates.get(countryiso, default_date)
            if date < max_date:
                continue
            country_dates[countryiso] = date
            closures[countryiso] = row["Status"]
        self.fully_closed = self.get_fully_closed(closures)
        for countryiso in closures:
            for region in self.regionlookup.iso3_to_region_and_hrp[countryiso]:
                if countryiso in self.fully_closed:
                    closed_countries[region] = (
                        closed_countries.get(region, 0) + 1
                    )

    def run_after_fallbacks(self) -> None:
        self.fully_closed = self.get_fully_closed(
            self.get_values("national")[0]
        )



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.aggregator import Aggregator
from hdx.scraper.runner import Runner

from .conftest import run_check_scraper, run_check_scrapers


class TestScrapersAggregation:
    def test_get_aggregation_hxl(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "MMR")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        name = "population"
        headers = (["Population"], ["#population"])
        national_values = [{"AFG": 38041754, "MMR": 54045420}]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            national_values,
            sources,
            population_lookup=national_values[0],
            source_urls=[
                "https://data.humdata.org/organization/world-bank-group"
            ],
        )

        regions = ("ROAP",)
        adm_aggregation = {"AFG": regions, "MMR": regions}

        configuration_hxl = configuration["aggregation_hxl"]
        level = "regional"
        aggregator_configuration = configuration_hxl["aggregation_sum"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        scraper_names = runner.add_customs(scrapers, add_to_run=True)
        name = "population_regional"
        assert scraper_names == [name]

        values = [{"ROAP": 92087174}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | values[0],
            source_urls=list(),
        )

        aggregator_configuration = configuration_hxl["aggregation_mean"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        runner.add_customs(scrapers, add_to_run=True)

        level = "regional"
        values = [{"ROAP": 46043587}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | values[0],
            source_urls=list(),
        )

        aggregator_configuration = configuration_hxl["aggregation_range"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        runner.add_customs(scrapers, add_to_run=True)

        level = "regional"
        values = [{"ROAP": "38041754-54045420"}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            source_urls=list(),
        )

        level = "national"
        BaseScraper.population_lookup = dict()
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        names = ("population", "who_national")
        headers = (
            [
                "Population",
                "CasesPer100000",
                "DeathsPer100000",
            ],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
            ],
        )
        national_values = [
            {"AFG": 38041754, "PSE": 4685306},
            {"AFG": "96.99", "PSE": "362.43"},
            {"AFG": "3.41", "PSE": "1.98"},
        ]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        source_urls = [
            "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            "https://data.humdata.org/organization/world-bank-group",
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            national_values,
            sources,
            source_urls=source_urls,
        )

        level = "global"
        aggregator_configuration = configuration_hxl[f"aggregation_{level}"]
        adm_aggregation = ("AFG", "PSE")
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        names = runner.add_customs(scrapers, add_to_run=True)
        assert names == [
            f"population_{level}",
            f"affected_infected_per100000_{level}",
            f"affected_infected_perpop_{level}",
        ]
        pop = 42727060
        headers = (
            ["Population", "CasesPer100000", "CasesPerPopulation"],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+infected+perpop",
            ],
        )
        values = [{"value": pop}, {"value": "229.71"}, {"value": "0.5376"}]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | {"global": pop},
            source_urls=list(),
        )

    def test_get_aggregation_nohxl(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "national"
        configuration_nohxl = configuration["aggregation_nohxl"]
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "MMR")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        name = "population"
        headers = (["Population"], ["#population"])
        national_values = [{"AFG": 38041754, "MMR": 54045420}]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            national_values,
            sources,
            population_lookup=national_values[0],
            source_urls=[
                "https://data.humdata.org/organization/world-bank-group"
            ],
        )

        adm_aggregation = {"AFG": ("ROAP",), "MMR": ("ROAP",)}

        level = "regional"
        aggregator_configuration = configuration_nohxl["aggregation_sum"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
            use_hxl=False,
        )
        scraper_names = runner.add_customs(scrapers, add_to_run=True)
        name = "population_regional"
        assert scraper_names == [name]

        values = [{"ROAP": 92087174}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | values[0],
            source_urls=list(),
        )

        aggregator_configuration = configuration_nohxl["aggregation_mean"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
            use_hxl=False,
        )
        runner.add_customs(scrapers, add_to_run=True)

        level = "regional"
        values = [{"ROAP": 46043587}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | values[0],
            source_urls=list(),
        )

        aggregator_configuration = configuration_nohxl["aggregation_range"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
            use_hxl=False,
        )
        runner.add_customs(scrapers, add_to_run=True)

        level = "regional"
        values = [{"ROAP": "38041754-54045420"}]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            list(),
            source_urls=list(),
        )

        level = "national"
        BaseScraper.population_lookup = dict()
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        names = ("population", "who_national")
        headers = (
            [
                "Population",
                "CasesPer100000",
                "DeathsPer100000",
            ],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
            ],
        )
        national_values = [
            {"AFG": 38041754, "PSE": 4685306},
            {"AFG": "96.99", "PSE": "362.43"},
            {"AFG": "3.41", "PSE": "1.98"},
        ]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        source_urls = [
            "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            "https://data.humdata.org/organization/world-bank-group",
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            national_values,
            sources,
            source_urls=source_urls,
            set_not_run=False,
        )

        level = "regional"
        aggregator_configuration = configuration_nohxl["aggregation_global"]
        adm_aggregation = ("AFG", "PSE")
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
            use_hxl=False,
        )
        names = runner.add_customs(scrapers, add_to_run=True)
        assert names == [
            f"population_{level}",
            f"casesper100000_{level}",
            f"casesperpopulation_{level}",
        ]
        pop = 42727060
        headers = (
            ["Population", "CasesPer100000", "CasesPerPopulation"],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+infected+perpop",
            ],
        )
        values = [{"value": pop}, {"value": "229.71"}, {"value": "0.5376"}]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | {"allregions": pop},
            source_urls=source_urls,
        )



import logging
from typing import Dict

from hdx.utilities.dictandlist import dict_of_sets_add
from hdx.utilities.typehint import ListTuple

from .reader import Read

logger = logging.getLogger(__name__)


class RegionLookup:
    iso3_to_region = dict()
    iso3_to_regions = dict()
    regions = None

    @classmethod
    def load(
        cls,
        region_config: Dict,
        countryiso3s: ListTuple[str],
        additional_regions: Dict[str, ListTuple],
    ):
        _, iterator = Read.get_reader().read_hdx(
            region_config,
            file_prefix="regions",
        )
        ignore_regions = region_config.get("ignore", tuple())
        for region in additional_regions:
            if region in ignore_regions:
                continue
            cls.iso3_to_regions[region] = dict()
        regions = set()
        toplevel_region = region_config.get("toplevel_region")
        if toplevel_region:
            cls.iso3_to_regions[toplevel_region] = dict()
        for row in iterator:
            countryiso3 = row[region_config["iso3_header"]]
            if countryiso3 and countryiso3 in countryiso3s:
                region = row[region_config["region_header"]]
                if region in ignore_regions:
                    continue
                regions.add(region)
                cls.iso3_to_region[countryiso3] = region
                if toplevel_region:
                    dict_of_sets_add(
                        cls.iso3_to_regions[toplevel_region],
                        countryiso3,
                        region,
                    )
                for additional_region in additional_regions:
                    if additional_region in ignore_regions:
                        continue
                    if countryiso3 in additional_regions[additional_region]:
                        dict_of_sets_add(
                            cls.iso3_to_regions[additional_region],
                            countryiso3,
                            region,
                        )
        cls.regions = sorted(list(regions))
        for additional_region in additional_regions:
            if additional_region in ignore_regions:
                continue
            cls.regions.insert(0, additional_region)
            for countryiso3 in additional_regions[additional_region]:
                for region in cls.iso3_to_regions:
                    dict_of_sets_add(
                        cls.iso3_to_regions[region],
                        countryiso3,
                        additional_region,
                    )
        cls.regions.insert(0, toplevel_region)
        for countryiso3 in countryiso3s:
            for region in cls.iso3_to_regions:
                dict_of_sets_add(
                    cls.iso3_to_regions[region], countryiso3, toplevel_region
                )



def get_fallbacks(fallback_data, level):
    data = fallback_data[f"{level}_data"]
    sources = fallback_data["sources"]
    if level == "national":
        admin_hxltag = "#country+code"
    else:
        admin_hxltag = "global"
    fallbacks = {
        "data": data,
        "admin hxltag": admin_hxltag,
        "sources": sources,
        "sources hxltags": [
            "#indicator+name",
            "#date",
            "#meta+source",
            "#meta+url",
        ],
    }
    return fallbacks


def bool_assert(actual, expected, msg):
    assert actual is expected, msg



from hdx.scraper.utilities.region_lookup import RegionLookup


class TestRegionLookup:
    def test_regionlookup(self, configuration):
        gho_countries = configuration["gho"]
        hrp_countries = configuration["HRPs"]

        regional_configuration = configuration["regional"]
        RegionLookup.load(
            regional_configuration, gho_countries, {"HRPs": hrp_countries}
        )
        assert RegionLookup.iso3_to_region["BRA"] == "ROLAC"
        assert RegionLookup.iso3_to_region["PAK"] == "ROAP"
        assert RegionLookup.iso3_to_region["UKR"] == "ROCCA"
        assert RegionLookup.iso3_to_regions["GHO"]["PAK"] == {
            "HRPs",
            "GHO",
            "ROAP",
        }
        assert RegionLookup.iso3_to_regions["HRPs"]["UKR"] == {
            "HRPs",
            "GHO",
            "ROCCA",
        }
        assert RegionLookup.iso3_to_regions["GHO"]["UKR"] == {
            "HRPs",
            "GHO",
            "ROCCA",
        }
        RegionLookup.iso3_to_region = dict()
        RegionLookup.iso3_to_regions = dict()

        regional_configuration = configuration["regional_ignore"]
        RegionLookup.load(
            regional_configuration, gho_countries, {"HRPs": hrp_countries}
        )
        assert RegionLookup.iso3_to_region["BRA"] == "ROLAC"
        assert RegionLookup.iso3_to_region["PAK"] == "ROAP"
        assert RegionLookup.iso3_to_region.get("UKR") is None
        assert RegionLookup.iso3_to_regions["GHO"]["PAK"] == {"GHO", "ROAP"}
        assert RegionLookup.iso3_to_regions.get("HRPs") is None
        assert RegionLookup.iso3_to_regions["GHO"]["UKR"] == {"GHO"}
        RegionLookup.iso3_to_region = dict()
        RegionLookup.iso3_to_regions = dict()



from datetime import datetime

from hdx.data.dataset import Dataset

from hdx.scraper.utilities import (
    get_isodate_from_dataset_date,
    string_params_to_dict,
)


class TestUtils:
    def test_string_params_to_dict(self):
        result = string_params_to_dict("a: 123, b: 345")
        assert result == {"a": "123", "b": "345"}
        result = string_params_to_dict("a:123,b:345")
        assert result == {"a": "123", "b": "345"}

    def test_get_isodate_from_dataset_date(self, configuration):
        dataset = Dataset(
            {
                "dataset_date": "[2022-01-11T02:24:08.241 TO 2022-01-11T02:24:08.241]"
            }
        )
        result = get_isodate_from_dataset_date(dataset, datetime.now())
        assert result == "2022-01-11"



from datetime import datetime

import pytest
from hdx.data.dataset import Dataset
from hdx.data.resource import Resource
from hdx.utilities.dateparse import parse_date
from hdx.utilities.downloader import Download
from hdx.utilities.path import temp_dir

from hdx.scraper.utilities.reader import Read


class TestReaders:
    def test_clone(self):
        with Download(user_agent="test") as downloader:
            with Read(
                downloader,
                "fallback_dir",
                "saved_dir",
                "temp_dir",
                save=False,
                use_saved=False,
                prefix="population",
                today=parse_date("2021-02-01"),
            ) as reader:
                clone_reader = reader.clone(downloader)
                for property, value in vars(reader).items():
                    if property == "downloader":
                        continue
                    assert getattr(clone_reader, property) == value

    def test_read_dataset(self, configuration, monkeypatch):
        def test_read_from_hdx(dataset_name):
            if dataset_name == "None":
                return None
            dataset = Dataset({"name": dataset_name})
            resource = Resource(
                {
                    "name": "test",
                    "url": "https://docs.google.com/spreadsheets/d/1NjSI2LaS3SqbgYc0HdD8oIb7lofGtiHgoKKATCpwVdY/edit#gid=1088874596",
                }
            )
            resource.set_file_type("csv")
            dataset.add_update_resource(resource)
            return dataset

        with temp_dir("TestReader") as temp_folder:
            with Download(user_agent="test") as downloader:
                munged_url = "https://docs.google.com/spreadsheets/d/1NjSI2LaS3SqbgYc0HdD8oIb7lofGtiHgoKKATCpwVdY/export?format=csv&gid=1088874596"
                with Read(
                    downloader,
                    temp_folder,
                    temp_folder,
                    temp_folder,
                    save=True,
                    use_saved=False,
                    prefix="test",
                    today=parse_date("2021-02-01"),
                ) as reader:
                    monkeypatch.setattr(
                        Dataset, "read_from_hdx", test_read_from_hdx
                    )
                    dataset_name = "None"
                    dataset = reader.read_dataset(dataset_name)
                    assert dataset is None
                    dataset_name = "Test Dataset"
                    dataset = reader.read_dataset(dataset_name)
                    assert dataset["name"] == dataset_name
                    assert dataset.get_resource()["url"] == munged_url
                    monkeypatch.delattr(Dataset, "read_from_hdx")
                with Read(
                    downloader,
                    temp_folder,
                    temp_folder,
                    temp_folder,
                    save=False,
                    use_saved=True,
                    prefix="test",
                    today=parse_date("2021-02-01"),
                ) as reader:
                    dataset_name = "None"
                    dataset = reader.read_dataset(dataset_name)
                    assert dataset is None
                    dataset_name = "Test Dataset"
                    dataset = reader.read_dataset(dataset_name)
                    assert dataset["name"] == dataset_name
                    assert dataset.get_resource()["url"] == munged_url

    def test_read_hxl_resource(self, fixtures):
        with temp_dir("TestReader") as temp_folder:
            with Download(user_agent="test") as downloader:
                with Read(
                    downloader,
                    temp_folder,
                    fixtures,
                    temp_folder,
                    save=False,
                    use_saved=True,
                    prefix="whowhatwhere",
                    today=parse_date("2021-02-01"),
                ) as reader:
                    resource = Resource(
                        {
                            "name": "3w data",
                            "url": "https://docs.google.com/spreadsheets/d/1NjSI2LaS3SqbgYc0HdD8oIb7lofGtiHgoKKATCpwVdY/edit#gid=1088874596",
                        }
                    )
                    resource.set_file_type("csv")
                    data = reader.read_hxl_resource("AFG", resource, "3w data")
                    assert len(data.headers) == 15
                    data = reader.read_hxl_resource(
                        "notags", resource, "3w data"
                    )
                    assert data is None
                    with pytest.raises(FileNotFoundError):
                        reader.read_hxl_resource(
                            "not exist", resource, "3w data"
                        )

    def test_read(self, configuration):
        url = Read.get_url("http://{{var}}", var="hello")
        assert url == "http://hello"
        datasetinfo = {
            "name": "test",
            "dataset": "sahel-humanitarian-needs-overview",
            "format": "csv",
        }
        reader = Read.get_reader()
        headers, iterator = reader.read(datasetinfo, file_type="csv")
        assert headers == [
            "Country",
            "nutrition",
            "Affected 2017",
            "In Need 2017",
            "Targeted 2017",
            "% targeted",
        ]
        assert next(iterator) == {
            "Country": "#country",
            "nutrition": "#sector?",
            "Affected 2017": "#affected",
            "In Need 2017": "#inneed",
            "Targeted 2017": "#targeted",
            "% targeted": "#targeted+percentage",
        }
        assert next(iterator) == {
            "Country": "Burkina Faso",
            "nutrition": "MAM",
            "Affected 2017": "433,412",
            "In Need 2017": "433,412",
            "Targeted 2017": "             _",
            "% targeted": "0",
        }
        date = datetime(2016, 9, 1, 23, 59, 59)
        assert datasetinfo == {
            "name": "test",
            "dataset": "sahel-humanitarian-needs-overview",
            "format": "csv",
            "headers": 1,
            "source_date": date,
            "source": "Multiple organisations",
            "source_url": "https://data.humdata.org/dataset/sahel-humanitarian-needs-overview",
            "url": "https://data.humdata.org/dataset/47f6ef46-500f-421a-9fa2-fefd93facf95/resource/2527ac5b-66fe-46f0-8b9b-7086d2c4ddd3/download/hno-2017-sahel-nutrition.csv",
        }
        datasetinfo = {
            "name": "test",
            "dataset": "sahel-humanitarian-needs-overview",
            "resource": "HNO-2017-Sahel- People in need.xlsx",
            "format": "xlsx",
            "sheet": 1,
        }
        headers, iterator = reader.read(datasetinfo)
        assert headers == [
            "Country",
            "Sector",
            "People in need",
            "Total population",
        ]
        assert next(iterator) == {
            "Country": "#country",
            "Sector": "#sector",
            "People in need": "#inneed",
            "Total population": "#total",
        }
        assert next(iterator) == {
            "Country": "Mali",
            "Sector": "Shelter/NFI",
            "People in need": 317000,
            "Total population": 100000,
        }
        assert datasetinfo == {
            "name": "test",
            "dataset": "sahel-humanitarian-needs-overview",
            "resource": "HNO-2017-Sahel- People in need.xlsx",
            "format": "xlsx",
            "sheet": 1,
            "headers": 1,
            "source_date": date,
            "source": "Multiple organisations",
            "source_url": "https://data.humdata.org/dataset/sahel-humanitarian-needs-overview",
            "url": "https://data.humdata.org/dataset/47f6ef46-500f-421a-9fa2-fefd93facf95/resource/d9248be4-7bfb-4a81-a7aa-c035dcb737a2/download/hno-2017-sahel-people-in-need.xlsx",
        }
        with pytest.raises(ValueError):
            datasetinfo = {"name": "test", "format": "unknown"}
            reader.read(datasetinfo)
        with pytest.raises(ValueError):
            datasetinfo = {
                "name": "test",
                "dataset": "sahel-humanitarian-needs-overview",
                "format": "json",
            }
            reader.read(datasetinfo)



import logging
from typing import List, Tuple

from hdx.utilities.loader import LoadError, load_json

logger = logging.getLogger(__name__)


class Fallbacks:
    fallbacks = None
    default_levels_mapping = {
        "global": "global_data",
        "regional": "regional_data",
        "national": "national_data",
        "subnational": "subnational_data",
    }
    default_admin_name_mapping = {
        "global": "value",
        "regional": "#region+name",
        "national": "#country+code",
        "subnational": "#adm1+code",
    }

    @classmethod
    def add(
        cls,
        fallbacks_path,
        levels_mapping=default_levels_mapping,
        sources_key="sources",
        admin_name_mapping=default_admin_name_mapping,
    ):
        try:
            fallback_data = load_json(fallbacks_path)
            fallback_sources = fallback_data[sources_key]
            sources_hxltags = [
                "#indicator+name",
                "#date",
                "#meta+source",
                "#meta+url",
            ]
            cls.fallbacks = dict()
            for level, output_key in levels_mapping.items():
                cls.fallbacks[level] = {
                    "data": fallback_data[output_key],
                    "admin name": admin_name_mapping[level],
                    "sources": fallback_sources,
                    "sources hxltags": sources_hxltags,
                }
        except (IOError, LoadError):
            cls.fallbacks = None

    @classmethod
    def exist(cls):
        return cls.fallbacks is not None

    @classmethod
    def get(
        cls,
        level,
        headers: Tuple[List, List],
    ) -> Tuple[List, List]:
        """Use provided fallbacks when there is a problem obtaining the latest data. The
        fallbacks dictionary should have the following keys: "data" containing a list of
        dictionaries from HXL hashtag to value, "admin name" to specify a particular admin
        name to use or "admin hxltag" specifying the HXL hashtag of the admin unit,
        "sources" containing a list of dictionaries with source information and
        "sources hxltags" containing a list of HXL hashtags with the name one first. eg.

        {"data": [{"#country+code": "AFG", "": "#value+wb+total": "572000000", ...}, ...],
        "admin hxltag": "#country+code",
        "sources": [{"#date": "2020-07-29", "#indicator+name": "#value+wb+total",
        "#meta+source": "OCHA, Center for Disaster Protection",
        "#meta+url": "https://data.humdata.org/dataset/compilation..."}, ...],
        "sources hxltags": ["#indicator+name", "#date", "#meta+source", "#meta+url"]}

        Args:
            headers (Tuple[List, List]): Headers

        Returns:
            Tuple[List, List]: Tuple of (Output values, output sources)
        """
        values = list()
        sources = list()

        if cls.fallbacks:
            fallbacks = cls.fallbacks[level]
            fb_data = fallbacks["data"]
            fb_adm_name = fallbacks.get("admin name", None)
            fb_adm_hxltag = fallbacks.get("admin hxltag", None)

            output_hxltags = headers[1]
            valdicts = [dict() for _ in output_hxltags]
            for row in fb_data:
                if fb_adm_name:
                    adm_key = fb_adm_name
                elif fb_adm_hxltag:
                    adm_key = fb_adm_hxltag
                else:
                    raise ValueError(
                        "Either admin name or admin hxltag must be specified!"
                    )
                if adm_key != "value":
                    adm_key = row[adm_key]
                for i, hxltag in enumerate(output_hxltags):
                    val = row.get(hxltag)
                    if val is not None:
                        valdicts[i][adm_key] = val
            values.extend(valdicts)
            fb_sources_hxltags = fallbacks["sources hxltags"]
            for row in fallbacks["sources"]:
                hxltag = row[fb_sources_hxltags[0]]
                if hxltag in output_hxltags:
                    sources.append(
                        (
                            hxltag,
                            row[fb_sources_hxltags[1]],
                            row[fb_sources_hxltags[2]],
                            row[fb_sources_hxltags[3]],
                        )
                    )
        return values, sources



import logging
from collections.abc import MutableMapping
from datetime import datetime
from os.path import join
from typing import Any, Dict, Iterator, List, Optional, Tuple
from urllib.parse import parse_qsl

import hxl
from hdx.data.dataset import Dataset
from hdx.data.resource import Resource
from hdx.utilities.dateparse import parse_date
from hdx.utilities.downloader import Download
from hdx.utilities.retriever import Retrieve
from hdx.utilities.saver import save_json
from hdx.utilities.typehint import ListTuple
from hxl.input import InputOptions, _munge_url
from slugify import slugify

from . import get_date_from_dataset_date, match_template

logger = logging.getLogger(__name__)


class Read(Retrieve):
    """Read data from tabular source eg. csv, xls, xlsx

    Args:
        downloader (Download): Download object
        fallback_dir (str): Directory containing static fallback data
        saved_dir (str): Directory to save or load downloaded data
        temp_dir (str): Temporary directory for when data is not needed after downloading
        save (bool): Whether to save downloaded data. Defaults to False.
        use_saved (bool): Whether to use saved data. Defaults to False.
        prefix (str): Prefix to add to filenames. Defaults to "".
        delete (bool): Whether to delete saved_dir if save is True. Defaults to True.
        today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).
    """

    def __init__(
        self,
        downloader: Download,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        prefix: str = "",
        delete: bool = True,
        today: Optional[datetime] = None,
    ):
        super().__init__(
            downloader,
            fallback_dir,
            saved_dir,
            temp_dir,
            save,
            use_saved,
            prefix,
            delete,
        )
        self.today: Optional[datetime] = today

    @classmethod
    def create_readers(
        cls,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        ignore: ListTuple[str] = tuple(),
        rate_limit: Optional[Dict] = {"calls": 1, "period": 0.1},
        today: Optional[datetime] = None,
        **kwargs: Any,
    ):
        """Generate a default reader and an HDX reader. Additional readers are generated
        if any of header_auths, basic_auths or extra_params are populated. header_auths
        and basic_auths are dictionaries of form {"scraper name": "auth", ...}.
        extra_params is of form {"scraper name": {"key": "auth", ...}, ...}.

        Args:
            fallback_dir (str): Directory containing static fallback data
            saved_dir (str): Directory to save or load downloaded data
            temp_dir (str): Temporary directory for when data is not needed after downloading
            save (bool): Whether to save downloaded data. Defaults to False.
            use_saved (bool): Whether to use saved data. Defaults to False.
            ignore (ListTuple[str]): Don't generate retrievers for these downloaders
            rate_limit (Optional[Dict]): Rate limiting per host. Defaults to {"calls": 1, "period": 0.1}
            today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).
            **kwargs: See below and parameters of Download class in HDX Python Utilities
            hdx_auth (str): HDX API key
            header_auths (Mapping[str, str]): Header authorisations
            basic_auths (Mapping[str, str]): Basic authorisations
            param_auths (Mapping[str, str]): Extra parameter authorisations

        Returns:
            None
        """
        if rate_limit:
            kwargs["rate_limit"] = rate_limit
        custom_configs = dict()
        hdx_auth = kwargs.get("hdx_auth")
        if hdx_auth:
            custom_configs["hdx"] = {"headers": {"Authorization": hdx_auth}}
            del kwargs["hdx_auth"]
        header_auths = kwargs.get("header_auths")
        if header_auths is not None:
            for name in header_auths:
                custom_configs[name] = {
                    "headers": {"Authorization": header_auths[name]}
                }
            del kwargs["header_auths"]
        basic_auths = kwargs.get("basic_auths")
        if basic_auths is not None:
            for name in basic_auths:
                custom_configs[name] = {"basic_auth": basic_auths[name]}
            del kwargs["basic_auths"]
        param_auths = kwargs.get("param_auths")
        if param_auths is not None:
            for name in param_auths:
                custom_configs[name] = {
                    "extra_params_dict": dict(parse_qsl(param_auths[name]))
                }
            del kwargs["param_auths"]
        Download.generate_downloaders(custom_configs, **kwargs)
        cls.generate_retrievers(
            fallback_dir,
            saved_dir,
            temp_dir,
            save,
            use_saved,
            ignore,
            today=today,
        )

    @classmethod
    def get_reader(cls, name: Optional[str] = None) -> "Read":
        """Get a generated reader given a name. If name is not supplied, the default
        one will be returned.

        Args:
            name (Optional[str]): Name of reader. Defaults to None (get default).

        Returns:
            Retriever: Reader object
        """
        return cls.get_retriever(name)

    @staticmethod
    def get_url(url: str, **kwargs: Any) -> str:
        """Get url from a string replacing any template arguments

        Args:
            url (str): Url to read
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            str: Url with any template arguments replaced
        """
        for kwarg in kwargs:
            exec(f'{kwarg}="{kwargs[kwarg]}"')
        template_string, match_string = match_template(url)
        if template_string:
            replace_string = eval(match_string)
            url = url.replace(template_string, replace_string)
        return url

    def clone(self, downloader: Download) -> "Read":
        """Clone a given reader but use the given downloader

        Args:
            downloader (Download): Downloader to use

        Returns:
            Read: Cloned reader

        """
        return Read(
            downloader,
            fallback_dir=self.fallback_dir,
            saved_dir=self.saved_dir,
            temp_dir=self.temp_dir,
            save=self.save,
            use_saved=self.use_saved,
            prefix=self.prefix,
            delete=False,
            today=self.today,
        )

    def read_tabular(
        self, datasetinfo: MutableMapping, **kwargs: Any
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data from tabular source eg. csv, xls, xlsx

        Args:
            datasetinfo (MutableMapping): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        url = self.get_url(datasetinfo["url"], **kwargs)
        sheet = datasetinfo.get("sheet")
        headers = datasetinfo.get("headers")
        if headers is None:
            headers = 1
            datasetinfo["headers"] = 1
        kwargs["headers"] = headers
        if isinstance(headers, list):
            kwargs["fill_merged_cells"] = True
        format = datasetinfo["format"]
        kwargs["format"] = format
        if not sheet and format in ("xls", "xlsx"):
            sheet = 1
        if sheet:
            kwargs["sheet"] = sheet
        compression = datasetinfo.get("compression")
        if compression:
            kwargs["compression"] = compression
        return self.get_tabular_rows(
            url,
            dict_form=True,
            **kwargs,
        )

    def read_dataset(self, dataset_name: str) -> Optional[Dataset]:
        """Read HDX dataset

        Args:
            dataset_name (str): Dataset name

        Returns:
            Optional[Dataset]: The dataset that was read or None
        """
        saved_path = join(self.saved_dir, f"{dataset_name}.json")
        if self.use_saved:
            logger.info(f"Using saved dataset {dataset_name} in {saved_path}")
            dataset = Dataset.load_from_json(saved_path)
        else:
            dataset = Dataset.read_from_hdx(dataset_name)
            if self.save:
                logger.info(f"Saving dataset {dataset_name} in {saved_path}")
                if dataset is None:
                    save_json(None, saved_path)
                else:
                    dataset.save_to_json(saved_path, follow_urls=True)
        return dataset

    def read_resource(self, identifier: str, resource: Resource):
        filename = f"{identifier}_{resource['name'].lower()}"
        file_type = f".{resource.get_file_type()}"
        if filename.endswith(file_type):
            filename = filename[: -len(file_type)]
        filename = f"{slugify(filename, separator='_')}{file_type}"
        url = _munge_url(resource["url"], InputOptions())
        path = self.download_file(url, filename=filename)
        return url, path

    def read_hxl_resource(
        self, identifier: str, resource: Resource, data_type: str
    ) -> Optional[hxl.Dataset]:
        try:
            _, path = self.read_resource(identifier, resource)
            data = hxl.data(path, InputOptions(allow_local=True)).cache()
            data.display_tags
            return data
        except hxl.HXLException:
            logger.warning(
                f"Could not process {data_type} for {identifier}. Maybe there are no HXL tags?"
            )
            return None
        except Exception:
            logger.exception(f"Error reading {data_type} for {identifier}!")
            raise

    def read_hdx_metadata(
        self, datasetinfo: MutableMapping
    ) -> Optional[Resource]:
        """Read metadata from HDX dataset and add to input dictionary

        Args:
            datasetinfo (MutableMapping): Dictionary of information about dataset

        Returns:
            Optional[Resource]: The resource if a url was not given
        """
        dataset_name = datasetinfo["dataset"]
        dataset = self.read_dataset(dataset_name)
        resource = None
        url = datasetinfo.get("url")
        if not url:
            resource_name = datasetinfo.get("resource")
            format = datasetinfo["format"].lower()
            for resource in dataset.get_resources():
                if resource["format"].lower() == format:
                    if resource_name and resource["name"] != resource_name:
                        continue
                    url = resource["url"]
                    break
            if not url:
                raise ValueError(
                    f"Cannot find {format} resource in {dataset_name}!"
                )
            datasetinfo["url"] = url
        date = datasetinfo.get("source_date")
        if date:
            if isinstance(date, str):
                datasetinfo["source_date"] = parse_date(date)
        else:
            datasetinfo["source_date"] = get_date_from_dataset_date(
                dataset, today=self.today
            )
        if "source" not in datasetinfo:
            datasetinfo["source"] = dataset["dataset_source"]
        if "source_url" not in datasetinfo:
            datasetinfo["source_url"] = dataset.get_hdx_url()
        return resource

    def read_hdx(
        self,
        datasetinfo: MutableMapping,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data and metadata from HDX dataset

        Args:
            datasetinfo (MutableMapping): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        self.read_hdx_metadata(datasetinfo)
        return self.read_tabular(datasetinfo, **kwargs)

    def read(
        self,
        datasetinfo: MutableMapping,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data and metadata from HDX dataset

        Args:
            datasetinfo (MutableMapping): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments in urls

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        format = datasetinfo["format"]
        if format in ["json", "csv", "xls", "xlsx"]:
            if "dataset" in datasetinfo:
                headers, iterator = self.read_hdx(datasetinfo, **kwargs)
            else:
                headers, iterator = self.read_tabular(datasetinfo, **kwargs)
        else:
            raise ValueError(
                f"Invalid format {format} for {datasetinfo['name']}!"
            )
        return headers, iterator



import copy
import logging
from datetime import datetime
from operator import itemgetter
from typing import Dict, Generator, Iterator, List, Optional, Tuple

import hxl
from hdx.location.adminone import AdminOne
from hdx.location.country import Country
from hdx.utilities.dateparse import parse_date
from hdx.utilities.dictandlist import dict_of_lists_add

from ..utilities import match_template

logger = logging.getLogger(__name__)


class RowParser:
    """RowParser class for parsing each row.

    Args:
        name (str): Name of scraper
        countryiso3s (List[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library that handles processing of admin level_type 1
        level (str): Can be national, subnational or single
        datelevel (str): Can be global, regional, national, subnational
        datasetinfo (Dict): Dictionary of information about dataset
        headers (List[str]): Row headers
        header_to_hxltag (Optional[Dict[str, str]]): Mapping from headers to HXL hashtags or None
        subsets (List[Dict]): List of subset definitions
        maxdateonly (bool): Whether to only take the most recent date. Defaults to True.
    """

    def __init__(
        self,
        name: str,
        countryiso3s: List[str],
        adminone: AdminOne,
        level: str,
        datelevel: str,
        datasetinfo: Dict,
        headers: List[str],
        header_to_hxltag: Optional[Dict[str, str]],
        subsets: List[Dict],
        maxdateonly: bool = True,
    ) -> None:
        def get_level(level: str) -> Optional[int]:
            """Get the level_name as a number. "Single" valued outputs are typically
            regional or global

            Args:
                level (str): Can be national, subnational or single (for a single value)

            Returns:
                Optional[int]: Level as a number
            """
            if level == "single":
                return None
            elif level == "national":
                return 0
            else:
                return 1

        self.name = name
        self.level = get_level(level)
        self.datelevel = get_level(datelevel)
        self.today = datasetinfo["source_date"]
        self.sort = datasetinfo.get("sort")
        self.stop_row = datasetinfo.get("stop_row")
        self.datecol = datasetinfo.get("date")
        self.datetype = datasetinfo.get("date_type")
        if self.datetype:
            if self.datetype == "date":
                date = parse_date("1900-01-01")
            else:
                date = 0
        else:
            date = 0
        self.maxdate = date
        self.single_maxdate = datasetinfo.get("single_maxdate", False)
        self.ignore_future_date = datasetinfo.get("ignore_future_date", True)
        self.adminone = adminone
        self.admcols = datasetinfo.get("admin", list())
        self.admexact = datasetinfo.get("admin_exact", False)
        self.admsingle = datasetinfo.get("admin_single", None)
        if self.admsingle:
            self.datelevel = None
        self.subsets = subsets
        self.filter_cols = datasetinfo.get("filter_cols", list())
        prefilter = datasetinfo.get("prefilter")
        if prefilter is not None:
            prefilter = self.get_filter_str_for_eval(prefilter)
        self.prefilter = prefilter
        adms = datasetinfo.get("admin_filter")
        if adms is None:
            self.adms = [countryiso3s, self.adminone.pcodes]
        else:
            if self.datelevel == 1:
                self.adms = adms
            else:
                self.adms = [adms, self.adminone.pcodes]
        if self.datelevel is None:
            self.maxdates = {i: date for i, _ in enumerate(subsets)}
        else:
            if self.datelevel > len(self.admcols):
                raise ValueError(
                    "No admin columns specified for required level_type!"
                )
            self.maxdates = {
                i: {adm: date for adm in self.adms[self.datelevel]}
                for i, _ in enumerate(subsets)
            }

        self.maxdateonly = maxdateonly
        self.flatteninfo = datasetinfo.get("flatten")
        self.headers = headers
        self.header_to_hxltag: Optional[Dict[str, str]] = header_to_hxltag
        self.filters = dict()
        self.read_external_filter(datasetinfo.get("external_filter"))

    def read_external_filter(self, external_filter: Optional[Dict]) -> None:
        """Read filter list from external url pointing to a HXLated file

        Args:
            external_filter (Optional[Dict]): External filter information in dictionary

        Returns:
            None
        """
        if not external_filter:
            return
        hxltags = external_filter["hxl"]
        data = hxl.data(external_filter["url"])
        for row in data:
            for hxltag in data.columns:
                if hxltag.display_tag in hxltags:
                    if self.header_to_hxltag:
                        header = hxltag.display_tag
                    else:
                        header = hxltag.header
                    dict_of_lists_add(
                        self.filters, header, row.get("#country+code")
                    )

    def get_filter_str_for_eval(self, filter):
        if self.filter_cols:
            for col in self.filter_cols:
                filter = filter.replace(col, f"row['{col}']")
        else:
            if self.datecol:
                filter = filter.replace(self.datecol, f"row['{self.datecol}']")
            for subset in self.subsets:
                for col in subset["input"]:
                    filter = filter.replace(col, f"row['{col}']")
        return filter

    def filter_sort_rows(self, iterator: Iterator[Dict]) -> Iterator[Dict]:
        """Apply prefilter and sort the input data before processing. If date_col is
        specified along with any of sum or process, and sorting is not specified, then
        apply a sort by date to ensure correct results.

        Args:
            iterator (Iterator[Dict]): Input data
        Returns:
            Iterator[Dict]: Input data with prefilter applied if specified and sorted if specified or deemed necessary
        """
        rows = list()
        for row in iterator:
            if self.header_to_hxltag:
                newrow = dict()
                for header in row:
                    newrow[self.header_to_hxltag[header]] = row[header]
                row = newrow
            if self.stop_row:
                if all(
                    row[key] == value for key, value in self.stop_row.items()
                ):
                    break
            for newrow in self.flatten(row):
                rows.append(newrow)
        if not self.sort:
            if self.datecol:
                for subset in self.subsets:
                    apply_sort = subset.get(
                        "sum",
                        subset.get("process", subset.get("input_append")),
                    )
                    if apply_sort:
                        logger.warning(
                            "sum or process used without sorting. Applying sort by date to ensure correct results!"
                        )
                        self.sort = {"keys": [self.datecol], "reverse": True}
                        break
        if self.prefilter:
            rows = [row for row in rows if eval(self.prefilter)]
        if self.sort:
            keys = self.sort["keys"]
            reverse = self.sort.get("reverse", False)
            rows = sorted(list(rows), key=itemgetter(*keys), reverse=reverse)
        return rows

    def flatten(self, row: Dict) -> Generator[Dict, None, None]:
        """Flatten a wide spreadsheet format into a long one

        Args:
            row (Dict): Row to flatten

        Returns:
            Generator[Dict]: Flattened row(s)
        """
        if not self.flatteninfo:
            yield row
            return
        counters = [-1 for _ in self.flatteninfo]
        while True:
            newrow = copy.deepcopy(row)
            for i, flatten in enumerate(self.flatteninfo):
                colname = flatten["original"]
                template_string, replace_string = match_template(colname)
                if not template_string:
                    raise ValueError(
                        "Column name for flattening lacks an incrementing number!"
                    )
                if counters[i] == -1:
                    counters[i] = int(replace_string)
                else:
                    replace_string = f"{counters[i]}"
                colname = colname.replace(template_string, replace_string)
                if colname not in row:
                    return
                newrow[flatten["new"]] = row[colname]
                extracol = flatten.get("extracol")
                if extracol:
                    newrow[extracol] = colname
                counters[i] += 1
            yield newrow

    def get_maxdate(self) -> datetime:
        """Get the most recent date of the rows so far

        Returns:
            datetime: Most recent date in processed rows
        """
        return self.maxdate

    def filtered(self, row: Dict) -> bool:
        """Check if the row should be filtered out

        Args:
            row (Dict): Row to check for filters

        Returns:
            bool: Whether row is filtered out or not
        """
        for header in self.filters:
            if header not in row:
                continue
            if row[header] not in self.filters[header]:
                return True
        return False

    def parse(self, row: Dict) -> Tuple[Optional[str], Optional[List[bool]]]:
        """Parse row checking for valid admin information and if the row should be filtered out in each subset given
        its definition.

        Args:
            row (Dict): Row to parse

        Returns:
            Tuple[Optional[str], Optional[List[bool]]]: (admin name, should process subset list) or (None, None)
        """
        if self.filtered(row):
            return None, None

        adms = [None for _ in range(len(self.admcols))]

        def get_adm(admcol, i):
            template_string, match_string = match_template(admcol)
            if template_string and self.headers:
                admcol = self.headers[int(match_string)]
            adm = row[admcol]
            if not adm:
                return False
            adm = adm.strip()
            adms[i] = adm
            if adm in self.adms[i]:
                return True
            exact = False
            if self.admexact:
                adms[i] = None
            else:
                if i == 0:
                    adms[i], exact = Country.get_iso3_country_code_fuzzy(adm)
                elif i == 1:
                    adms[i], exact = self.adminone.get_pcode(
                        adms[0], adm, self.name
                    )
                if adms[i] not in self.adms[i]:
                    adms[i] = None
            return exact

        for i, admcol in enumerate(self.admcols):
            if admcol is None:
                continue
            if isinstance(admcol, str):
                admcol = [admcol]
            for admcl in admcol:
                exact = get_adm(admcl, i)
                if adms[i] and exact:
                    break
            if not adms[i]:
                return None, None

        should_process_subset = list()
        for subset in self.subsets:
            filter = subset["filter"]
            process = True
            if filter:
                filter = self.get_filter_str_for_eval(filter)
                if not eval(filter):
                    process = False
            should_process_subset.append(process)

        if self.datecol:
            if isinstance(self.datecol, list):
                dates = [str(row[x]) for x in self.datecol]
                date = "".join(dates)
            else:
                date = row[self.datecol]
            if self.datetype == "date":
                if not isinstance(date, datetime):
                    date = parse_date(date)
                date = date.replace(tzinfo=None)
                if date > self.today and self.ignore_future_date:
                    return None, None
            elif self.datetype == "year":
                date = int(date)
                if date > self.today.year and self.ignore_future_date:
                    return None, None
            else:
                date = int(date)
            for i, process in enumerate(should_process_subset):
                if not process:
                    continue
                if date < self.maxdate:
                    if self.single_maxdate:
                        should_process_subset[i] = False
                else:
                    self.maxdate = date
                if self.datelevel is None:
                    if self.maxdateonly:
                        if date < self.maxdates[i]:
                            should_process_subset[i] = False
                        else:
                            self.maxdates[i] = date
                    else:
                        self.maxdates[i] = date
                else:
                    if self.maxdateonly:
                        if date < self.maxdates[i][adms[self.datelevel]]:
                            should_process_subset[i] = False
                        else:
                            self.maxdates[i][adms[self.datelevel]] = date
                    else:
                        self.maxdates[i][adms[self.datelevel]] = date
        if self.level is None:
            return "value", should_process_subset
        if self.admsingle:
            return self.admsingle, should_process_subset
        return adms[self.level], should_process_subset



from typing import Any, Dict, List, Optional, Union

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None


class BaseOutput:
    """Base class for output that can also be used for testing as it does nothing.

    Args:
        updatetabs (List[str]): Tabs to update
    """

    def __init__(self, updatetabs: List[str]) -> None:
        self.updatetabs = updatetabs

    def update_tab(
        self,
        tabname: str,
        values: Union[List, DataFrame],
        hxltags: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        """Update tab with values. Classes that inherit from this one should
        implement this method.

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Values in a list of lists or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.
            **kwargs

        Returns:
            None
        """
        return

    def add_data_row(self, key: str, row: Dict) -> None:
        """Add row

        Args:
            key (str): Key to update
            row (Dict): Row to add

        Returns:
            None
        """
        return

    def add_dataframe_rows(
        self, key: str, df: DataFrame, hxltags: Optional[Dict] = None
    ) -> None:
        """Add rows from dataframe under a key

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        return

    def add_data_rows_by_key(
        self,
        name: str,
        countryiso: str,
        rows: List[Dict],
        hxltags: Optional[Dict] = None,
    ) -> None:
        """Add rows under both a key and an ISO 3 country code subkey

        Args:
            key (str): Key to update
            countryiso (str): Country to use as subkey
            rows (List[Dict]): List of dictionaries
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        return

    def add_additional(self) -> None:
        """Download files and add them under keys defined in the configuration

        Returns:
            None
        """
        return

    def save(self, **kwargs: Any) -> None:
        """Save file

        Args:
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            None
        """
        return



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.aggregator import Aggregator
from hdx.scraper.runner import Runner

from .conftest import run_check_scrapers


class TestRunnerGetResults:
    def test_get_results(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "MMR")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        level = "national"
        BaseScraper.population_lookup = dict()
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, adminone, today)
        runner.add_configurables(scraper_configuration, level)

        names = ("population", "who_national")
        headers = (
            [
                "Population",
                "CasesPer100000",
                "DeathsPer100000",
            ],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
            ],
        )
        national_values = [
            {"AFG": 38041754, "PSE": 4685306},
            {"AFG": "96.99", "PSE": "362.43"},
            {"AFG": "3.41", "PSE": "1.98"},
        ]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "2020-08-06",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        source_urls = [
            "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            "https://data.humdata.org/organization/world-bank-group",
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            national_values,
            sources,
            source_urls=source_urls,
            set_not_run=False,
        )

        level = "global"
        configuration_hxl = configuration["aggregation_hxl"]
        aggregator_configuration = configuration_hxl[f"aggregation_{level}"]
        adm_aggregation = ("AFG", "PSE")
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            runner,
        )
        names = runner.add_customs(scrapers, add_to_run=True)
        assert names == [
            f"population_{level}",
            f"affected_infected_per100000_{level}",
            f"affected_infected_perpop_{level}",
        ]
        pop = 42727060
        headers = (
            ["Population", "CasesPer100000", "CasesPerPopulation"],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+infected+perpop",
            ],
        )
        values = [{"value": pop}, {"value": "229.71"}, {"value": "0.5376"}]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            list(),
            population_lookup=national_values[0] | {"global": pop},
            source_urls=source_urls,
            set_not_run=False,
        )

        regions = ("ROAP",)
        aggregator_configuration = configuration_hxl["aggregation_sum"]
        scrapers = Aggregator.get_scrapers(
            aggregator_configuration,
            "national",
            "regional",
            {"AFG": regions, "PSE": ("somewhere",)},
            runner,
        )
        regional_names = runner.add_customs(scrapers, add_to_run=True)
        runner.run(regional_names)
        results = runner.get_results()
        assert results == {
            "national": {
                "headers": (
                    ["Population", "CasesPer100000", "DeathsPer100000"],
                    [
                        "#population",
                        "#affected+infected+per100000",
                        "#affected+killed+per100000",
                    ],
                ),
                "values": [
                    {"AFG": 38041754, "PSE": 4685306},
                    {"AFG": "96.99", "PSE": "362.43"},
                    {"AFG": "3.41", "PSE": "1.98"},
                ],
                "sources": [
                    (
                        "#population",
                        "2020-10-01",
                        "World Bank",
                        "https://data.humdata.org/organization/world-bank-group",
                    ),
                    (
                        "#affected+infected+per100000",
                        "2020-08-06",
                        "WHO",
                        "https://covid19.who.int/WHO-COVID-19-global-data.csv",
                    ),
                    (
                        "#affected+killed+per100000",
                        "2020-08-06",
                        "WHO",
                        "https://covid19.who.int/WHO-COVID-19-global-data.csv",
                    ),
                ],
                "fallbacks": [],
            },
            "global": {
                "headers": (
                    ["Population", "CasesPer100000", "CasesPerPopulation"],
                    [
                        "#population",
                        "#affected+infected+per100000",
                        "#affected+infected+perpop",
                    ],
                ),
                "values": [
                    {"value": 42727060},
                    {"value": "229.71"},
                    {"value": "0.5376"},
                ],
                "sources": [],
                "fallbacks": [],
            },
            "regional": {
                "headers": (["Population"], ["#population"]),
                "values": [{"ROAP": 38041754, "somewhere": 4685306}],
                "sources": [],
                "fallbacks": [],
            },
        }
        results = runner.get_results(
            names=("population_global",), levels=("national", "global")
        )
        assert results == {
            "global": {
                "headers": (["Population"], ["#population"]),
                "values": [{"value": 42727060}],
                "sources": [],
                "fallbacks": [],
            }
        }
        results = runner.get_results(
            names=("population_global",),
            levels=("world", "global"),
            overrides={
                "population_global": {"national": "world", "global": "world"}
            },
        )
        assert results == {
            "world": {
                "headers": (["Population"], ["#population"]),
                "values": [{"value": 42727060}],
                "sources": [],
                "fallbacks": [],
            }
        }
